#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:class:`QAbstractSpinBox`-based simulation configuration widget subclasses.
'''

#FIXME: Improve the init() method to internally call the self.setMinimum()
#and/or self.setMaximum() methods with a range of permissible values specific to
#the current "self._sim_conf_alias" data descriptor. Although the expr_alias()
#function currently provides no means of explicitly defining this range, it
#*DOES* provide a means of passing a predicate callable which could actually be
#a callable class (i.e., class defining the __call__() special method)
#explicitly defining and exposing this range as public instance variables.
#Specifically, we might define the following predicate classes:
#
#    class PredicateUnaryABC(meta=ABC):
#        '''
#        **Unary predicate** (i.e., callable class whose :meth:`__call__`
#        special method accepts a single value to be tested and returns ``True``
#        only if this value satisfies this predicate).
#        '''
#
#        @abstractmethod
#        def __call__(self, value: object) -> bool:
#            pass
#
#
#    class PredicateUnaryRange(PredicateUnaryABC):
#
#        @type_check
#        def __init__(
#            self, min: NumericOrNoneTypes, max: NumericOrNoneTypes) -> None:
#            self.min = min
#            self.max = max
#
#        def __call__(self, value: object) -> bool:
#            return (
#               (self.min is None or value >= self.min) and
#               (self.max is None or value <= self.max)
#            )
#
#The expr_alias() function would then need to additionally expose the passed
#predicate (if any) by declaring a new instance variable "expr_alias_predicate"
#on the instance of the class returned by that function.
#
#Given that, the init() method could then internally call:
#
#    predicate = self._sim_conf_alias.data_desc.predicate
#
#    if isinstance(predicate, PredicateUnaryRange):
#        if predicate.min is not None:
#            self.setMinimum(predicate.min)
#        if predicate.max is not None:
#            self.setMinimum(predicate.max)
#
#It's *NOT* hard. It's mostly just tedious. But it *MUST* be done to prevent
#users from entering invalid numeric data. (Let us see this through, please.)

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, Qt, Signal
from PySide2.QtWidgets import QSpinBox
# from betse.util.io.log import logs
from betse.util.type.numeric import floats
from betse.util.type.types import type_check, ClassOrNoneTypes
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditscalar import (
    QBetseeSimConfEditScalarWidgetMixin)
from betsee.util.widget.abc.guiclipboardabc import (
    QBetseeClipboardScalarWidgetMixin)
from betsee.util.widget.stock.guispinbox import QBetseeDoubleSpinBox

# ....................{ SUPERCLASSES                       }....................
class QBetseeSimConfSpinBoxWidgetMixin(
    QBetseeClipboardScalarWidgetMixin, QBetseeSimConfEditScalarWidgetMixin):
    '''
    Abstract base class of all simulation configuration-specific subclasses,
    permitting numeric values (i.e., integers, floating point values) backed by
    external simulation configuration files to be interactively edited.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Enable so-called "acceleration," both increasing and decreasing the
        # rate at which the displayed number changes as a function of the
        # duration of time that the spon box GUI arrows are depressed.
        self.setAccelerated(True)

        # Align the displayed number against the right rather than left internal
        # edge of this spin box -- the typical default in most GUI frameworks.
        # (For unclear reasons, Qt defaults to right alignment.)
        self.setAlignment(Qt.AlignRight)

        # Disable so-called "keyboard tracking," reducing signal verbosity by
        # preventing the superclass from emitting valueChanged() signals on each
        # key entered by the user (e.g., on all three of the "3", "1", and "4"
        # keys entered by the user to enter the integer "314").
        #
        # Instead, the valueChanged() signal will *ONLY* be emitted when:
        #
        # * The return key is pressed.
        # * Keyboard focus is lost.
        # * Other spinbox functionality is activated (e.g., an up or down arrow
        #   key is pressed, the up or down graphical arrow is clicked).
        self.setKeyboardTracking(False)

    # ..................{ SUPERCLASS ~ setter                }..................
    def setValue(self, value_new: str) -> None:

        # logs.log_debug('In QBetseeSimConfSpinBoxWidgetMixin.setValue()...')

        # Defer to the superclass setter.
        super().setValue(value_new)

        # If this configuration is currently open, set the current value of this
        # simulation configuration alias to this widget's current value.
        self._set_alias_to_widget_value_if_sim_conf_open()

    # ..................{ MIXIN ~ property                   }..................
    @property
    def undo_synopsis(self) -> str:
        return QCoreApplication.translate(
            'QBetseeSimConfSpinBoxWidgetMixin', 'edits to a spin box')


    @property
    def widget_value(self) -> object:
        return self.value()


    @property
    def _finalize_widget_change_signal(self) -> Signal:
        return self.editingFinished

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfIntSpinBox(
    QBetseeSimConfSpinBoxWidgetMixin, QSpinBox):
    '''
    Simulation configuration-specific integer spin box widget, permitting
    integers backed by external simulation configuration files to be
    interactively edited.
    '''

    # ..................{ MIXIN                              }..................
    @property
    def _sim_conf_alias_type_strict(self) -> ClassOrNoneTypes:
        return int


    @QBetseeSimConfSpinBoxWidgetMixin.widget_value.setter
    @type_check
    def widget_value(self, widget_value: int) -> None:

        # logs.log_debug('In QBetseeSimConfIntSpinBox.widget_value()...')

        # Set this widget's displayed value to the passed value by calling the
        # setValue() method of our superclass rather than this subclass,
        # preventing infinite recursion. (See the superclass method docstring.)
        QSpinBox.setValue(self, widget_value)


    def _reset_widget_value(self) -> None:
        self.widget_value = 0


class QBetseeSimConfDoubleSpinBox(
    QBetseeSimConfSpinBoxWidgetMixin, QBetseeDoubleSpinBox):
    '''
    Simulation configuration-specific floating point spin box widget, permitting
    floating point numbers backed by external simulation configuration files to
    be interactively edited.
    '''

    # ..................{ MIXIN                              }..................
    @property
    def _sim_conf_alias_type_strict(self) -> ClassOrNoneTypes:
        return float


    # This method is typically called *ONLY* once on loading the current
    # simulation configuration, enabling this widget to reinitialize itself in a
    # manner dependent upon the current value of the associated alias.
    def _get_widget_from_alias_value(self) -> object:

        # Initial value of the floating point number returned by the simulation
        # configuration alias associated with this widget.
        widget_value = super()._get_widget_from_alias_value()

        # Base-10 precision (i.e., significand length) of this initial value.
        widget_value_precision = floats.get_base_10_precision(widget_value)

        # If this widget defers to the default single step (i.e., fractional
        # value by which to increment and decrement this widget's displayed
        # value on each push of this widget's up and down arrows) for
        # "QDoubleSpinBox" widgets...
        if self.singleStep() == self.SINGLE_STEP_DEFAULT:
            # Base-10 exponent of this initial value.
            widget_value_exponent = floats.get_base_10_exponent(widget_value)

            # Override the default single step with a widget-specific single
            # step incrementing and decrementing by a fraction of this initial
            # value. Specifically, a single step of a quarter of this initial
            # value's order of magnitude is set.
            #
            # For example, given initial values of:
            #
            # * "0.009", a single step of "0.00025" will be set.
            # * "1.729", a single step of "0.25" will be set.
            single_step_new = 2.5 * 10**(widget_value_exponent - 1)
            self.setSingleStep(single_step_new)
        # Else, this widget has a non-default single step (typically specified
        # at Qt Creator time). In this case, preserve this single step as is.

        # Refine this precision to the largest of:
        widget_value_precision = max(
            # This alias' precision, incremented by one to permit the end user
            # to interactively decrease this number an additional decimal place.
            widget_value_precision + 1,
            # A reasonable default precision.
            3,
        )

        # Set the precision of this widget's displayed value to this precision.
        # logs.log_debug(
        #     'Setting "%s" precision given %s to %d...',
        #     self.obj_name, widget_value, widget_value_precision)
        self.setDecimals(widget_value_precision)

        # Return this value.
        return widget_value


    @QBetseeSimConfSpinBoxWidgetMixin.widget_value.setter
    @type_check
    def widget_value(self, widget_value: float) -> None:

        # logs.log_debug('In QBetseeSimConfDoubleSpinBox.widget_value()...')

        # Set this widget's displayed value to the passed value by calling the
        # setValue() method of our superclass rather than this subclass,
        # preventing infinite recursion. (See the superclass method docstring.)
        QBetseeDoubleSpinBox.setValue(self, widget_value)


    def _reset_widget_value(self) -> None:
        self.widget_value = 0.0



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`PySide2`-specific widget facilities universally applicable to
all (or at least most) widget types.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific submodules of this subpackage (i.e.,
# "betsee.util.widget"). Since those submodules must *ALWAYS* be able to safely
# import from this submodule, circularities are best avoided here.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from PySide2.QtWidgets import QWidget
# from betse.util.io.log import logs
from betse.util.type.types import type_check, CallableTypes
from betsee.guiexception import BetseePySideWidgetException
from betsee.util.type.guitype import QWidgetOrNoneTypes

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_widget_parent_satisfies(
    widget: QWidget, predicate: CallableTypes) -> None:
    '''
    Raise an exception unless some transitive parent widget of the passed widget
    satisfies the passed predicate.

    Equivalently, this function raises an exception if *no* transitive parent
    widget of the passed widget satisfies the passed predicate.

    Parameters
    ----------
    widget : QWidgetOrNoneTypes
        Child widget to be inspected.
    predicate : CallableTypes
        Callable accepting one passed parent widget, returning ``True`` only if
        this widget satisfies this predicate. Hence, this predicate should have
        a signature resembling: ``def predicate(widget: QWidget) -> bool``.

    Raises
    ----------
    BetseePySideWidgetException
        If *no* transitive parent widget of this child widget satisfies this
        boolean predicate.
    '''

    if not is_widget_parent_satisfies(widget=widget, predicate=predicate):
        raise BetseePySideWidgetException(
            'Parent widget of child widget "{0}" '
            'satsifying predicate {1!r} not found.'.format(
                widget.objectName(), predicate))

# ....................{ TESTERS                            }....................
@type_check
def is_widget_parent_satisfies(
    widget: QWidget, predicate: CallableTypes) -> bool:
    '''
    ``True`` only if some transitive parent widget of the passed widget
    satisfies the passed predicate.

    For each transitive parent widget of the passed widget in ascending order
    from the immediate to root parent widget (e.g., :class:`QMainWindow` widget)
    of the passed widget, this function iteratively:

    #. Passes that parent widget to the passed predicate.
    #. If that predicate returns ``True``, this function halts searching and
       immediately returns ``True``.
    #. Else if that parent widget itself has a parent widget, this function
       continues searching by iterating up the widget ownership hierarchy to
       that parent parent widget and repeating the above logic.
    #. Else, returns ``False``.

    Parameters
    ----------
    widget : QWidgetOrNoneTypes
        Child widget to be inspected.
    predicate : CallableTypes
        Callable accepting one passed parent widget, returning ``True`` only if
        this widget satisfies this predicate. Hence, this predicate should have
        a signature resembling: ``def predicate(widget: QWidget) -> bool``.

    Returns
    ----------
    bool
        ``True`` only if some parent of this widget satisfies this predicate.
    '''

    # Currently visited child widget, starting at the passed child widget.
    widget_child = widget

    # Currently visited parent widget if this child widget has a parent widget
    # (i.e., is not the root widget) *OR* "None" otherwise.
    widget_parent = None

    # For here to eternity...
    while True:
        # Currently visited parent widget if any *OR* "None" otherwise.
        widget_parent = widget_child.parentWidget()

        # If this parent satisfies this predicate, halt searching with success.
        if predicate(widget_parent):
            return True

        # Else, this parent does *NOT* satisfy this predicate. In this case,
        # continue searching by iterating into the parent of this parent widget.
        widget_child = widget_parent

    # If no parents satisfy this predicate, halt searching with failure.
    return False

# ....................{ GETTERS                            }....................
@type_check
def get_label(widget: QWidgetOrNoneTypes) -> str:
    '''
    Human-readable label synopsizing the passed widget if any.

    Parameters
    ----------
    widget : QWidgetOrNoneTypes
        Widget to be synopsized *or* ``None``, in which case the absence of a
        widget is synopsized.

    Returns
    ----------
    str
        Human-readable label synopsizing this widget if any.
    '''

    return (
        'widget "{}"'.format(widget.objectName()) if widget is not None else
        'no widget')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`QMessageBox`-based message handling facilities.

See Also
----------
:mod:`betsee.util.io.guierr`
    Error-specific :class:`QMessageBox` facilities.
'''

#FIXME: Move into the "betsee.util.io" submodule.

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QMessageBox
from betse.util.type.types import type_check, StrOrNoneTypes
from betsee.guiexception import BetseePySideMessageBoxException

# ....................{ SHOWERS ~ severity                 }....................
def show_query(*args, **kwargs) -> int:
    '''
    Display the passed question(s) as a :mod:`QMessageBox`-driven modal message
    box in the current application widget.

    See Also
    ----------
    :func:`show_message`
        Further details.
    '''

    return show_message(*args, severity=QMessageBox.Question, **kwargs)


def show_info(*args, **kwargs) -> int:
    '''
    Display the passed information as a :mod:`QMessageBox`-driven modal message
    box in the current application widget.

    See Also
    ----------
    :func:`show_message`
        Further details.
    '''

    return show_message(*args, severity=QMessageBox.Information, **kwargs)


def show_warning(*args, **kwargs) -> int:
    '''
    Display the passed warning message(s) as a :mod:`QMessageBox`-driven modal
    message box in the current application widget.

    See Also
    ----------
    :func:`show_message`
        Further details.
    '''

    return show_message(*args, severity=QMessageBox.Warning, **kwargs)

# ....................{ SHOWERS ~ message                  }....................
@type_check
def show_message(
    # Mandatory parameters.
    title: str,
    synopsis: str,

    # Optional parameters.
    severity: QMessageBox.Icon = QMessageBox.NoIcon,
    buttons: QMessageBox.StandardButtons = QMessageBox.Ok,
    button_default: QMessageBox.StandardButton = QMessageBox.Ok,
    exegesis: StrOrNoneTypes = None,
    details: StrOrNoneTypes = None,
) -> int:
    '''
    Display the passed warning message(s) as a :mod:`PySide2`-driven modal
    message box in the current application widget.

    Parameters
    ----------
    title : str
        Title of this message box.
    synopsis : str
        Synopsis (i.e., main text) of this of this message box.
    severity: optional[QMessageBox.Icon]
        Integer value of the :class:`QMessageBox.` enumeration member signifying
        the severity of the icon displayed by this message box. Defaults to
        :attr:`QMessageBox.NoIcon`, preventing an icon from being displayed.
    buttons : optional[QMessageBox.StandardButtons]
        Bit flag of the integer values of all OR-ed
        :class:`QMessageBox.StandardButton` enumeration members comprising the
        set of all buttons to be displayed by this message box. Defaults to
        :attr:`QMessageBox.Ok`, the conventional "OK" button.
    button_default : optional[QMessageBox.StandardButton]
        Bit value of the :class:`QMessageBox.StandardButton` enumeration member
        signifying the **default button** (i.e., button initially receiving the
        keyboard focus) displayed by this message box. If the passed ``buttons``
        bit field does *not* enable this bit, an exception is raised. Defaults
        to :attr:`QMessageBox.Ok`, the conventional "OK" button.
    exegesis : optional[str]
        Exegesis (i.e., explanation) of this warning to be displayed as the
        so-called "informative text" of this message box below the synopsis of
        this warning. Defaults to ``None``, in which case no such text is
        displayed.
    details : optional[str]
        Technical details of this warning to be displayed as the so-called
        "detailed text" of this message box in monospaced font below both the
        synopsis and exegesis of this warning in a discrete fold-down text area.
        Defaults to ``None``, in which case no such text is displayed.

    Returns
    ----------
    int
        Bit value of the :class:`QMessageBox.StandardButton` enumeration member
        signifying the button clicked by the user if any.
    '''

    # If the "buttons" bit field is empty, raise an exception.
    if buttons == 0:
        raise BetseePySideMessageBoxException(QCoreApplication.translate(
            'show_message', 'Button bit field empty.'))

    # If the "buttons" bit field does *NOT* enable this bit, raise an exception.
    if button_default & buttons == 0:
        raise BetseePySideMessageBoxException(QCoreApplication.translate(
            'show_message',
            'Default button bit "{0}" not enabled by '
            'button bit field "{1}".'.format(button_default, buttons)))

    # Message box displaying this warning.
    message_box = QMessageBox()
    message_box.setWindowTitle(title)
    message_box.setText(synopsis)
    message_box.setIcon(severity)
    message_box.setStandardButtons(buttons)
    message_box.setDefaultButton(button_default)

    # If this exception provides optional metadata, display this metadata.
    if exegesis is not None:
        message_box.setInformativeText(exegesis)
    if details is not None:
        message_box.setDetailedText(details)

    # Finalize this message box *AFTER* setting all widget proporties above.
    message_box.show()

    # Run this application's event loop, displaying this message box, and return
    # the bit value of the enumeration member signifying the button clicked by
    # the user if any.
    return message_box.exec_()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`setuptools`-based makefile instrumenting all high-level administration
tasks (e.g., installation, freezing, test running) for this application.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from packages guaranteed to exist at the start of
# installation. This includes all standard Python and application packages but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#
# Technically, this script may import from all subpackages and submodules of
# the this application's eponymous package. By Python mandate, the first
# element of the "sys.path" list is guaranteed to be the directory containing
# this script. Python necessarily searches this directory for imports from the
# local version of this application *BEFORE* any other directories (including
# system directories containing older versions of this application). To quote:
#
#     "As initialized upon program startup, the first item of this list,
#      path[0], is the directory containing the script that was used to invoke
#      the Python interpreter."
#
# See also: https://stackoverflow.com/a/10097543/2809027
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import setuptools
from betsee import guimetadata, guimetadeps
from betsee_setup import beupbuild, beuputil

# ....................{ METADATA                          }....................
# PyPI-specific metadata declared here rather than in the "betsee.metadata"
# submodule, as the main codebase neither requires nor desires this metadata.

# ....................{ METADATA ~ seo                    }....................
_KEYWORDS = ['biology', 'multiphysics', 'science', 'simulator',]
'''
List of all lowercase alphabetic keywords synopsising this application.

These keywords may be arbitrarily selected so as to pretend to improve search
engine optimization (SEO). In actuality, they do absolutely nothing.
'''


# To minimize desynchronization woes, all
# "Programming Language :: Python :: "-prefixed strings are dynamically
# appended to this list by the get_classifiers() function called below.
_CLASSIFIERS = [
    #FIXME: Replace with the following after the "1.0.0" release:
    #    'Development Status :: 5 - Production/Stable',

    # PyPI-specific version type. The number specified here is a magic constant
    # with no relation to this application's version numbering scheme. *sigh*
    'Development Status :: 4 - Beta',

    # Sublist of all supported platform-specific CLI and GUI components.
    'Environment :: MacOS X',
    'Environment :: Win32 (MS Windows)',
    'Environment :: X11 Applications',

    # Miscellaneous metadata.
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
]
'''
List of all PyPI-specific trove classifier strings synopsizing this
application.

Each such string *must* be contain either two or three `` :: `` substrings
delimiting human-readable capitalized English words formally recognized by the
:mod:`distutils`-specific ``register`` command.

See Also
----------
https://pypi.python.org/pypi?%3Aaction=list_classifiers
    Plaintext list of all trove classifier strings recognized by PyPI.
'''

# ....................{ OPTIONS                           }....................
# Setuptools-specific options. Keywords not explicitly recognized by either
# setuptools or distutils must be added to the above dictionary instead.
_SETUP_OPTIONS = {
    # ..................{ CORE                              }..................
    # Self-explanatory metadata.
    'name':             guimetadata.PACKAGE_NAME,
    'version':          guimetadata.VERSION,
    'author':           guimetadata.AUTHORS,
    'author_email':     guimetadata.AUTHOR_EMAIL,
    'maintainer':       guimetadata.AUTHORS,
    'maintainer_email': guimetadata.AUTHOR_EMAIL,
    'description':      guimetadata.SYNOPSIS,
    'long_description': beuputil.get_description(),
    'url':              guimetadata.URL_HOMEPAGE,
    'download_url':     guimetadata.URL_DOWNLOAD,

    # ..................{ PYPI                              }..................
    # PyPi-specific metadata.
    'classifiers': beuputil.sanitize_classifiers(_CLASSIFIERS),
    'keywords':    _KEYWORDS,
    'license':     guimetadata.LICENSE,

    # ..................{ DEPENDENCIES                      }..................
    # Mandatory nuntime dependencies.
    'install_requires': guimetadeps.get_runtime_mandatory_tuple(),

    # Optional nuntime dependencies. Whereas mandatory dependencies are defined
    # as sequences, optional dependencies are defined as a dictionary mapping
    # from an arbitrary alphanumeric word to a sequence containing one or more
    # such dependencies. Such dependencies are then installable via "pip" by
    # suffixing the name of this project by the "["- and "]"-delimited key
    # defined below whose value lists the dependencies to be installed (e.g.,
    # "sudo pip3 install betse[all]", installing both the application and all
    # mandatory and optional dependencies required by the application).
    'extras_require': {
        # All optional runtime dependencies.
        'all': guimetadeps.get_runtime_optional_tuple(),
    },

    # Mandatory testing dependencies.
    'tests_require': guimetadeps.get_testing_mandatory_tuple(),

    # ..................{ PACKAGES                           }..................
    # List of all Python packages (i.e., directories containing zero or more
    # Python modules) to be installed. Currently, this includes the "betse"
    # package and all subpackages of this package excluding:
    #
    # * The top-level test package and all subpackages of this package, test
    #   functionality *NOT* intended to be installed with this application.
    # * The top-level setup package and all subpackages of this package,
    #   setuptools functionality required only for application installation.
    # * "build", caching both setuptools metadata and a complete copy of this
    #   package, required only by a prior application installation.
    # * "freeze", providing PyInstaller-specific functionality required only for
    #   application freezing (i.e., conversion into an executable binary).
    'packages': setuptools.find_packages(exclude=(
        guimetadata.PACKAGE_NAME + '_test',
        guimetadata.PACKAGE_NAME + '_test.*',
        guimetadata.PACKAGE_NAME + '_setup',
        guimetadata.PACKAGE_NAME + '_setup.*',
        'build',
        'freeze',
    )),

    # ..................{ PATHS                              }..................
    # Cross-platform script wrappers dynamically created at installation time.
    'entry_points': {
        # GUI-specific scripts.
        'gui_scripts': (
            '{} = {}.__main__:main'.format(
                guimetadata.SCRIPT_BASENAME, guimetadata.PACKAGE_NAME),
        ),
    },

    # Install all data files (i.e., non-Python files) embedded in the Python
    # package tree for this application.
    #
    # Unlike Python packages, undesirable data files are excludable from
    # installation *ONLY* via the external "MANIFEST.in" file. This is
    # terrible, of course. (Did you expect otherwise?)
    #
    # Data files are *NOT* Python modules and hence should *NOT* be embedded in
    # the Python package tree. Sadly, the "data_files" key supported by
    # setuptools for this purpose is *NOT* cross-platform-portable and is thus
    # inherently broken. Why? Because this key either requires usage of absolute
    # paths *OR* relative paths relative to absolute paths defined by
    # "setup.cfg"; in either case, these paths are absolute. While the current
    # platform could be detected and the corresponding absolute path embedded in
    # 'data_files', that implementation would be inherently fragile. (That's
    # bad.) In lieu of sane setuptools support, we defer to the methodology
    # employed by everyone. Setuptools, your death is coming.
    'include_package_data': True,

    # Install to an uncompressed directory rather than a compressed archive.
    #
    # While nothing technically precludes the latter, doing so substantially
    # complicates runtime access of data files compressed into this archive
    # (e.g., with the pkg_resources.resource_filename() function). How so? By
    # decompressing this archive's contents into a temporary directory on
    # program startup and removing these contents on program shutdown. Since
    # there exists no guarantee this removal will actually be performed (e.g.,
    # due to preemptive SIGKILLs), compressed archives are inherently fragile.
    'zip_safe': False,

    # ..................{ COMMANDS                           }..................
    # Set of all custom setuptools subcommands specific to this makefile (e.g.,
    # "sudo python3 setup.py symlink"), defaulting to the empty set. Each
    # subsequent call to the add_setup_commands() function iteratively performed
    # below adds one or more such subcommands to this set.
    'cmdclass': {},
}
'''
Dictionary passed to the subsequent call to the :func:`setup` function.

This dictionary signifies the set of all application-specific :mod:`setuptools`
options. Submodules of the top-level :mod:`betsee_setup` package subsequently
customize these options (e.g., by defining custom commands).
'''
# print('extras: {}'.format(setup_options['extras_require']))


_SETUP_OPTIONS_CUSTOM = {
    # While currently empty, it's likely we'll want this again... someday.
}
'''
Non-setuptools-specific metadata, used to inform custom subcommands (e.g.,
``freeze_file``) of other metadata *not* already declared by the
:data:`setup_options` dictionary.

Setuptools raises fatal exceptions when the :data:`setup_options` dictionary
contains unrecognized keys. For safety, these keys are added to this dictionary
instead.
'''

# ....................{ COMMANDS                           }....................
# Define all application-specific setuptools commands.
for _setup_module in (beupbuild,):
    _setup_module.add_setup_commands(_SETUP_OPTIONS_CUSTOM, _SETUP_OPTIONS)

# ....................{ SETUP                              }....................
setuptools.setup(**_SETUP_OPTIONS)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Metadata constants synopsizing high-level application behaviour.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys

# ....................{ METADATA                          }....................
NAME = 'BETSEE'
'''
Human-readable application name.
'''


LICENSE = '2-clause BSD'
'''
Human-readable name of the license this application is licensed under.
'''

# ....................{ PYTHON ~ version                  }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Changes to this section *MUST* be synchronized with:
# * The corresponding section of the "betsee.metadata" submodule.
# * Front-facing documentation (e.g., "README.rst", "doc/md/INSTALL.md").
# On bumping the minimum required version of Python, consider also documenting
# the justification for doing so in the "Python Version" section of this
# submodule's docstring above.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

PYTHON_VERSION_MIN = '3.5.0'
'''
Human-readable minimum version of Python required by this application as a
``.``-delimited string.
'''


PYTHON_VERSION_MINOR_MAX = 8
'''
Maximum minor stable version of this major version of Python currently released
(e.g., ``5`` if Python 3.5 is the most recent stable version of Python 3.x).
'''


def _convert_version_str_to_tuple(version_str: str) -> tuple:
    '''
    Convert the passed human-readable ``.``-delimited version string into a
    machine-readable version tuple of corresponding integers.
    '''
    assert isinstance(version_str, str), (
        '"{}" not a version string.'.format(version_str))

    return tuple(
        int(version_part) for version_part in version_str.split('.'))


PYTHON_VERSION_MIN_PARTS = _convert_version_str_to_tuple(PYTHON_VERSION_MIN)
'''
Machine-readable minimum version of Python required by this application as a
tuple of integers.
'''


# Validate the version of the active Python interpreter *BEFORE* subsequent
# code possibly depending on such version. Since such version should be
# validated both at setuptools-based install time and post-install runtime
# *AND* since this module is imported sufficiently early by both, stash such
# validation here to avoid duplication of such logic and hence the hardcoded
# Python version.
#
# The "sys" module exposes three version-related constants for this purpose:
#
# * "hexversion", an integer intended to be specified in an obscure (albeit
#   both efficient and dependable) hexadecimal format: e.g.,
#    >>> sys.hexversion
#    33883376
#    >>> '%x' % sys.hexversion
#    '20504f0'
# * "version", a human-readable string: e.g.,
#    >>> sys.version
#    2.5.2 (r252:60911, Jul 31 2008, 17:28:52)
#    [GCC 4.2.3 (Ubuntu 4.2.3-2ubuntu7)]
# * "version_info", a tuple of three or more integers *OR* strings: e.g.,
#    >>> sys.version_info
#    (2, 5, 2, 'final', 0)
#
# For sanity, this application will *NEVER* conditionally depend upon the
# string-formatted release type of the current Python version exposed via the
# fourth element of the "version_info" tuple. Since the first three elements of
# that tuple are guaranteed to be integers *AND* since a comparable 3-tuple of
# integers is declared above, comparing the former and latter yield the
# simplest and most reliable Python version test.
#
# Note that the nearly decade-old and officially accepted PEP 345 proposed a
# new field "requires_python" configured via a key-value pair passed to the
# call to setup() in "setup.py" (e.g., "requires_python = ['>=2.2.1'],"), that
# field has yet to be integrated into either disutils or setuputils. Hence,
# that field is validated manually in the typical way.
if sys.version_info[:3] < PYTHON_VERSION_MIN_PARTS:
    # Human-readable current version of Python. "sys.version" is sufficiently
    # overly verbose as to be unusuable, sadly.
    PYTHON_VERSION = '.'.join(
        str(version_part) for version_part in sys.version_info[:3])

    # Die ignominiously.
    raise RuntimeError(
        '{} requires at least Python {}, but the active interpreter '
        'is only Python {}. We feel deep sadness for you.'.format(
            NAME, PYTHON_VERSION_MIN, PYTHON_VERSION))

# ....................{ METADATA ~ version                }....................
VERSION = '0.9.0.0'
'''
Human-readable application version as a ``.``-delimited string.

Design
----------
**This string must be prefixed by the minimum version of BETSE required by this
application.** Equivalently, this string must be of the form
``{betse_version}.{patch_version}`, where:

* ``{betse_version}`` is the minimum required version of BETSE.
* ``{patch_version}`` is the current patch number of this application with
  respect to this version of BETSE.

For example, if this is the second iteration of this application to require at
least BETSE 0.8.5, this version is expected to be ``0.8.5.2``.

This is *not* merely a style convention; this is a hard prerequisite. Why?
Because this string is parsed elsewhere (e.g., the `:mod:`betsee.guimetadeps`
submodule) to produce a setuptools-specific dependency on this BETSE version.
Failure to conform to this specification will induce user dependency hell at
both distribution and installation time, which is (rather) bad.
'''


VERSION_PARTS = _convert_version_str_to_tuple(VERSION)
'''
Machine-readable application version as a tuple of integers.
'''

# ....................{ METADATA ~ synopsis               }....................
# Note that a human-readable multiline description is exposed via the top-level
# "setup.py" script. This description is inefficiently culled from the contents
# of the top-level "README.rst" file and hence omitted here. (Doing so here
# would significantly increase program startup costs with little to no benefit.)
SYNOPSIS = 'BETSEE, the BioElectric Tissue Simulation Engine Environment.'
'''
Human-readable single-line synopsis of this application.

By PyPI design, this string must *not* span multiple lines or paragraphs.
'''


DESCRIPTION = (
    'The BioElectric Tissue Simulation Engine Environment (BETSEE) is the '
    'official Qt 5-based graphical user interface (GUI) for BETSE, a '
    'finite volume simulator for 2D computational multiphysics problems in '
    'the life sciences -- including electrodiffusion, electro-osmosis, '
    'galvanotaxis, voltage-gated ion channels, gene regulatory networks, '
    'and biochemical reaction networks.'
)
'''
Human-readable multiline description of this application.

By :mod:`argparse` design, this string may (and typically should) span both
multiple lines and paragraphs. Note that this string is *not* published to
PyPI, which accepts reStructuredText (rst) and is thus passed the contents of
the top-level :doc:`/README` file instead.
'''

# ....................{ METADATA ~ authors                }....................
AUTHORS = 'Alexis Pietak, Cecil Curry, et al.'
'''
Human-readable list of all principal authors of this application as a
comma-delimited string.

For brevity, this string *only* lists authors explicitly assigned copyrights.
For the list of all contributors regardless of copyright assignment or
attribution, see the top-level `AUTHORS.md` file.
'''


AUTHOR_EMAIL = 'leycec@gmail.com'
'''
Email address of the principal corresponding author (i.e., the principal author
responding to public correspondence).
'''

# ....................{ METADATA ~ organization           }....................
ORG_NAME = 'Paul Allen Discovery Center'
'''
Human-readable list of the single organization principally responsible for
funding the authors of this application.
'''


ORG_DOMAIN_NAME = 'alleninstitute.org'
'''
Machine-readable name of the top-level domain (TLD) hosting the organization
signified by the :data:`ORG_NAME` global.
'''

# ....................{ METADATA ~ urls                   }....................
URL_HOMEPAGE = 'https://gitlab.com/betse/betsee'
'''
URL of this application's homepage.
'''


URL_DOWNLOAD = (
    'https://gitlab.com/betse/betsee/repository/archive.tar.gz?ref=v{}'.format(
        VERSION,
    )
)
'''
URL of the source tarball for the current version of this application.

This URL assumes a tag whose name is ``v{VERSION}`` where ``{VERSION}`` is the
human-readable current version of this application (e.g., ``v0.4.0``) to exist.
Typically, no such tag exists for live versions of this application -- which
have yet to be stabilized and hence tagged. Hence, this URL is typically valid
*only* for previously released (rather than live) versions of this application.
'''

# ....................{ METADATA ~ python                 }....................
PACKAGE_NAME = NAME.lower()
'''
Fully-qualified name of the top-level Python package implementing this
application.
'''


SCRIPT_BASENAME = PACKAGE_NAME
'''
Basename of the CLI-specific Python script wrapper created by :mod:`setuptools`
installation.
'''

# ....................{ METADATA ~ python : main window   }....................
MAIN_WINDOW_QRC_MODULE_NAME = PACKAGE_NAME + '_rc'
'''
Fully-qualified name of the top-level Python package implementing this
this application's main window Qt resource collection (QRC) converted from the
corresponding XML-formatted UI file exported by the external Qt Designer GUI.

This module is dynamically generated at runtime and hence may *not* yet exist,
in which case the caller is assumed to safely generate this module before its
first importation.
'''


MAIN_WINDOW_UI_MODULE_NAME = PACKAGE_NAME + '_ui'
'''
Fully-qualified name of the top-level Python package implementing this
this application's main window user interface (UI) converted from the
corresponding XML-formatted UI file exported by the external Qt Designer GUI.

This module is dynamically generated at runtime and hence may *not* yet exist,
in which case the caller is assumed to safely generate this module before its
first importation.
'''

# ....................{ METADATA ~ private                }....................
_IS_TESTING = False
'''
``True`` only if the active Python interpreter is running a test session (e.g.,
with the ``py.test`` test harness).

This private global is subject to change and thus *not* intended to be publicly
accessed. Consider calling the public :func:`betse.util.py.pys.is_testing`
function instead.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **simulator enumeration** (e.g., :class:`enum.Enum` subclass
describing different types of simulator state) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.enums import make_enum
# from betse.util.type.types import type_check

# ....................{ ENUMERATIONS                      }....................
SimmerState = make_enum(
    class_name='SimmerState',
    member_names=(
        'UNQUEUED',
        'QUEUED',
        'MODELLING',
        'EXPORTING',
        'PAUSED',
        'STOPPING',
        'FINISHED',
    ),
    doc='''
    Enumeration of all supported types of **simulator state** (i.e., mutually
    exclusive combination of one or more booleans uniquely capturing the
    condition of the currently queued simulation subcommand if any, analogous
    to a state in a finite state automata).

    Attributes
    ----------
    UNQUEUED : enum
        Unqueued state, implying no subcommands to be queued.
    QUEUED : enum
        Queued state, implying one or more subcommands to be queued but *not*
        yet run and hence neither paused, halted, nor done.
    MODELLING : enum
        Modelling state, implying one or more queued subcommands specific to
        modelling (e.g., seed, initialization) to be currently running and
        hence neither paused, halted, nor done.
    EXPORTING : enum
        Exporting state, implying one or more queued subcommands specific to
        exporting (e.g., seed exports, initialization exports) to be currently
        running and hence neither paused, halted, nor done.
    PAUSED : enum
        Paused state, implying one or more queued subcommands to have been run
        but paused before completion.
    STOPPING : enum
        Stopping state, implying one or more queued subcommands to have been
        run but manually stopped before completion. This is a temporary state
        during which the proactor waits for the previously running worker to
        gracefully stop. Until this worker does so, the proactor remains
        incapable of performing new work (i.e., running queued subcommands).
        After this worker gracefully stops, the proactor switches from this
        state into the :attr:`FINISHED` state, implying the proactor to be
        capable of performing new work.
    FINISHED : enum
        Completion state, implying all queued subcommands to have completed
        either:

        * Successfully, in which case those subcommands ran to completion.
        * Unsuccessfully, in which case either:

          * The user prematurely stopped some or all of those subcommands.
          * One of those subcommands raised an uncaught fatal exception.
    '''
)



SimmerModelState = make_enum(
    class_name='SimmerModelState',
    member_names=(
        'PREPARING',
        'MODELLING',
        'FINISHING',
    ),
    doc='''
    Enumeration of all supported types of **simulator modelling state** (i.e.,
    mutually exclusive boolean uniquely capturing the condition of the
    currently queued modelling simulation subcommand if any, analogous to a
    state in a finite state automata).

    This enumeration applies *only* when the proactor is currently modelling
    (i.e., when the state of the proactor is :attr:`SimmerState.MODELLING`).
    This enumeration generically applies to modelling all simulation phases,
    including the seed, initialization, and simulation phases of a simulation.
    That said, this enumeration is typically only used to describe simulation
    phases whose modelling entails time steps -- notably, the initialization
    and simulation (but *not* seed) phases.

    Attributes
    ----------
    PREPARING : enum
        Pre-processing state. When modelling the initialization and simulation
        (but *not* seed) phases, this state implies a proactor worker to be
        preparing to model the first time step of this phase but to have yet to
        actually do so.
    MODELLING : enum
        In-processing state. When modelling the initialization and simulation
        (but *not* seed) phases, this state implies a proactor worker to be
        modelling a time step of this phase.
    FINISHING : enum
        Post-processing state, implying a proactor worker to:

        * Have successfully modelled all time steps of this phase.
        * Now be saving the results of this phase to disk, typically as one or
          more Python-specific pickled files.
    '''
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **stacked simulation configuration** (i.e., partitioning of the
simulation configuration into multiple pages, each displaying and editing all
settings associated with a single simulation feature  of the current such
configuration) facilities.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Slot  # QCoreApplication,
from PySide2.QtWidgets import QMainWindow, QStackedWidget, QTreeWidgetItem
from betse.util.io.log import logs
from betse.util.type.obj import objiter
from betse.util.type.text import strs
from betse.util.type.types import type_check
from betsee.gui.window.guinamespace import SIM_CONF_STACK_PAGE_NAME_PREFIX
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ CLASSES                            }....................
class QBetseeSimConfStackedWidget(QBetseeObjectMixin, QStackedWidget):
    '''
    :mod:`PySide2`-based stack widget containing multiple pages, each displaying
    and editing all settings associated with a single simulation feature (e.g.,
    ions, plots, tissue) of the current simulation configuration.

    This application-specific widget augments the stock :class:`QStackedWidget`
    with support for handling simulation configurations, including:

    * Integration with the corresponding :class:`QStackedWidget`, exposing all
      low-level configuration settings for the high-level simulation feature
      currently selected from this tree.

    Caveats
    ----------
    Each child page widget of this stack widget should have a name prefixed by a
    substring uniquely identifying this widget to be a child page and the name
    of the parent page widget of this child page widget if any. Specifically:

    * If this child page widget is top-level (i.e., has no parent), this is the
      :attr:`SIM_CONF_STACK_PAGE_NAME_PREFIX` substring.
    * If this child page widget is *not* top-level (i.e., has a parent), this is
      the name of that parent page widget followed by a delimiting underscore.

    As a corrolary, this implies the names of all page widgets regardless of
    hierarchical nesting to be prefixed by the
    :attr:`SIM_CONF_STACK_PAGE_NAME_PREFIX` substring. No other widgets should
    have such names. Failure to comply may be met with an unutterable anguish.

    Parameters
    ----------
    _pagers : tuple
        Container of all high-level objects controlling the state of each stack
        widget page, preserved to prevent Python from erroneously garbage
        collecting these objects. Since these objects are *NOT* explicitly
        accessed after instantiation, this simplistic scheme suffices.
    _sim_conf : QBetseeSimConf
        High-level object controlling simulation configuration state.
    _tree_item_to_stack_page : dict
        Dictionary mapping from each :class:`QTreeWidgetItem` in the
        :class:`QTreeWidget` associated with this stack widget to each child
        page widget of this stack widget. This dictionary ensures that clicking
        on each :class:`QTreeWidgetItem` displays the corresponding page widget.

    See Also
    ----------
    QBetseeSimConfTreeWidget
        Corresponding :class:`QTreeWidget` instance, exposing all high-level
        features of the current simulation configuration which this
        :class:`QStackedWidget` instance then exposes the low-level settings of.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._pagers = None
        self._sim_conf = None
        self._tree_item_to_stack_page = None

        # Instantiate all pages of this stack widget.
        self._init_pagers()


    @type_check
    def _init_pagers(self) -> None:
        '''
        Instantiate each **page controller** (i.e., high-level
        :mod:`PySide2`-based object encapsulating the internal state of a page
        of this stack widget).
        '''

        # Defer method-specific imports for maintainability.
        from betsee.gui.simconf.stack.pager.guisimconfpagerion import (
            QBetseeSimConfIonStackedWidgetPager)
        from betsee.gui.simconf.stack.pager.guisimconfpagerpath import (
            QBetseeSimConfPathStackedWidgetPager)
        from betsee.gui.simconf.stack.pager.guisimconfpagerspace import (
            QBetseeSimConfSpaceStackedWidgetPager)
        from betsee.gui.simconf.stack.pager.guisimconfpagertime import (
            QBetseeSimConfTimeStackedWidgetPager)
        from betsee.gui.simconf.stack.pager.guisimconfpagertis import (
            QBetseeSimConfTissueDefaultStackedWidgetPager)

        # Tuple of all stack widget page controllers defined in arbitrary order.
        self._pagers = (
            QBetseeSimConfIonStackedWidgetPager(),
            QBetseeSimConfPathStackedWidgetPager(),
            QBetseeSimConfSpaceStackedWidgetPager(),
            QBetseeSimConfTimeStackedWidgetPager(),
            QBetseeSimConfTissueDefaultStackedWidgetPager(),
        )


    # To avoid circular import dependencies, this parameter is validated to be
    # an instance of the "QMainWindow" superclass rather than the expected
    # "QBetseeMainWindow" subclass of the "betsee.gui.window.guimainwindow"
    # submodule. Why? Because the latter imports the cached "betsee_ui.py"
    # module which imports the current submodule. Since this application only
    # contains one main window, this current validation suffices.
    @type_check
    def init(self, main_window: QMainWindow) -> None:
        '''
        Initialize this stacked widget against the passed parent main window.

        This method is principally intended to perform **post-population
        initialization** (i.e., initialization performed *after* the main window
        has been completely pre-populated with all initial child widgets).

        To avoid circular references, this method is guaranteed to *not* retain
        a reference to this main window on returning. References to child
        widgets (e.g., simulation configuration stack widget) of this window may
        be retained, however.

        Parameters
        ----------
        main_window: QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this widget.
        '''

        # Initialize our superclass with all passed parameters.
        super().init()

        # Log this initialization.
        logs.log_debug('Initializing top-level stacked widget...')

        # Classify all variables accessed by subsequent slot invocations. Since
        # these variables do *NOT* own or otherwise preserve references to this
        # object, this is guaranteed to avoid circularities.
        self._sim_conf = main_window.sim_conf

        # Integrate this stack widget with this window's top-level tree widget.
        self._init_tree_to_stack(main_window)

        # Initialize each page of this stack widget.
        for pager in self._pagers:
            pager.init(main_window)

    # ..................{ INITIALIZERS ~ tree                }..................
    @type_check
    def _init_tree_to_stack(self, main_window: QMainWindow) -> None:
        '''
        Recursively map all tree widget items of the simulation
        configuration-specific tree widget in the passed parent main window to
        the corresponding child page widgets of this parent stack widget.

        Specifically, this method initializes the
        :attr:`_tree_item_to_stack_page` instance variable.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Parent :class:`QMainWindow` widget to initialize this widget with.
        '''

        # Initialize this dictionary.
        self._tree_item_to_stack_page = {}

        # Top-level tree widget associated with this stack widget.
        tree_widget = main_window.sim_conf_tree

        # Generator iteratively yielding a 2-tuple of the name and value of each
        # child page of this stack widget, matching all instance variables of
        # this main window with names prefixed by an identifying substring.
        stack_pages = objiter.iter_vars_custom_simple_prefixed(
            obj=main_window, prefix=SIM_CONF_STACK_PAGE_NAME_PREFIX)

        # Dictionary mapping the name to value of each such child page,
        # enabling iteration below to efficiently look these values up.
        stack_page_name_to_value = {
            stack_page_name: stack_page
            for stack_page_name, stack_page in stack_pages
        }

        # Recursion function intentionally defined as a closure to permit the
        # local variables defined above to be trivially and efficiently shared
        # between each recursive call of this function.
        @type_check
        def _map_tree_item_children_to_stack_pages(
            tree_item: QTreeWidgetItem,
            stack_page_name_prefix: str,
        ) -> None:
            '''
            Recursively map all child tree widget items of the passed parent
            tree widget item to the corresponding child page widgets of this
            parent stack widget.

            Parameters
            ----------
            tree_item : QTreeWidgetItem
                Parent tree widget item to recursively map all children of.
            stack_page_name_prefix : str
                Substring prefixing the names of all such child page widgets.
                If this parent tree widget item is:
                * The invisible placeholder root item, this is simply
                  :data:`SIM_CONF_STACK_PAGE_NAME_PREFIX`.
                * Any other item (including top-level, mid-level, and leaf
                  items) , this is the name of the child page widget
                  corresponding to this item suffixed by an underscore.
            '''

            # For the 0-based index of each child item of this parent item...
            #
            # Note that we intentionally iterate by tree widget items rather
            # than stack widget pages here. Why? Because iterating by the latter
            # would erroneously handle pages with no corresponding items, which
            # currently exist (e.g., temporary scratchpad purposes *NOT*
            # intended to be displayed to end users).
            for tree_item_child_index in range(tree_item.childCount()):
                # Current child item of this parent item.
                tree_item_child = tree_item.child(tree_item_child_index)

                # Text of this item in the first column.
                tree_item_child_text = tree_item_child.text(0)

                # Log this visitation.
                # logs.log_debug(
                #     'Visiting tree item "%s" child "%s"...',
                #     tree_item.text(), tree_item_child_text)

                # Name of the corresponding child page, synthesized from the
                # text of this child item as follows (in order):
                #
                # * Strip all whitespace from this child item's text (e.g., from
                #   "File Management" to "FileManagement").
                # * Prefix this child item's text by the identifying substring
                #   prefixing this child page's name (e.g., from
                #   "FileManagement" to 'sim_conf_stack_page_FileManagement').
                stack_page_name = (
                    stack_page_name_prefix +
                    strs.remove_whitespace(tree_item_child_text)
                )

                # Child page associated with this tree widget item if any or
                # "None" otherwise.
                stack_page = stack_page_name_to_value.get(stack_page_name, None)

                #FIXME: Convert to a fatal exception after finalizing this GUI.

                # If no such page exists...
                if stack_page is None:
                    # Log a non-fatal warning.
                    logs.log_warning(
                        'Simulation configuration-specific stacked page '
                        '"%s" not found.', stack_page_name)

                    # Skip to the next tree widget item.
                    continue

                    # raise BetseePySideTreeWidgetException(
                    #     QCoreApplication.translate(
                    #         'QBetseeSimConfStackedWidget',
                    #         'Simulation configuration-specific stacked page '
                    #         '"{0}" not found.'.format(stack_page_name)))

                # Log this recursion.
                logs.log_debug(
                    'Mapping tree item "%s" to stacked page "%s"...',
                    tree_item_child_text, stack_page_name)

                # Map this child item to this child page.
                self._tree_item_to_stack_page[tree_item_child] = stack_page

                # If this child item itself contains child items, recursively
                # map all child items of this child item.
                #
                # Note that this test constitutes a negligible optimization but
                # is otherwise unnecessary. Performing this call for a child
                # item with no child items reduces to an inefficient noop. That
                # said, since the call stack is limited in Python, we prefer to
                # avoid exhausting the stack if we can do so. (So, we do so.)
                if tree_item_child.childCount():
                    _map_tree_item_children_to_stack_pages(
                        tree_item=tree_item_child,

                        # Require the names of all child pages of this child page to
                        # be prefixed by the name of this child page delimited by an
                        # underscore, a simple convention that saves sanity.
                        stack_page_name_prefix=stack_page_name + '_',
                    )

        # Recursively map all tree widget items of this tree widget.
        _map_tree_item_children_to_stack_pages(
            # Start at the invisible placeholder root item, provided by Qt for
            # exactly this recursive purpose.
            tree_item=tree_widget.invisibleRootItem(),

            # Substring prefixing the names of *ALL* stack page widgets.
            stack_page_name_prefix=SIM_CONF_STACK_PAGE_NAME_PREFIX,
        )

    # ..................{ SLOTS ~ public                     }..................
    # The following public slots are connected to from other widgets.

    @Slot(QTreeWidgetItem, QTreeWidgetItem)
    def switch_page_to_tree_item(
        self,
        tree_item_curr: QTreeWidgetItem,
        tree_item_prev: QTreeWidgetItem,
    ) -> None:
        '''
        Switch to the child page widget of this stack widget associated with the
        passed tree widget item clicked by the end user.

        Parameters
        ----------
        tree_item_curr : QTreeWidgetItem
            Current tree widget item clicked by the end user.
        tree_item_prev : QTreeWidgetItem
            Previous tree widget item clicked by the end user.
        '''

        # Stack widget page associated with this tree widget item if any or
        # "None" otherwise.
        stack_page = self._tree_item_to_stack_page.get(tree_item_curr, None)

        #FIXME: Convert to a fatal exception after finalizing this GUI.

        # If no such page exists (e.g., due to this tree widget item being a
        # placeholder container for child items for which pages do exist)...
        if stack_page is None:
            # Log a non-fatal warning.
            logs.log_warning(
                'Simulation configuration-specific tree item "%s" '
                'associated with no stacked page.', tree_item_curr.text())

            # Ignore this attempt to switch the child page widget.
            return

        # Else, switch to this page.
        self.setCurrentWidget(stack_page)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

'''
Low-level :mod:`PySide2`-specific logging handler subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Signal
from betse.util.type.types import type_check
from logging import Handler

# ....................{ CLASSES                            }....................
#FIXME: Post as an answer to the following StackOverflow question:
#    https://stackoverflow.com/questions/14349563/how-to-get-non-blocking-real-time-behavior-from-python-logging-module-output-t
class LogHandlerSignal(Handler):
    '''
    :class:`Signal`-based handler, redirecting each log record sent to this
    handler to each slot connected to the signal with which this handler was
    initialized.

    Parameters
    ----------
    _signal : Signal
        Signal to redirect log records to.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, signal: Signal, *args, **kwargs) -> None:
        '''
        Initialize this handler to log with the passed signal.

        Parameters
        ----------
        signal : Signal
            Signal to redirect log records to.

        All remaining parameters are passed as is to our superclass method.
        '''

        # Initialize our superclass with all remaining passed parameters.
        super().__init__(*args, **kwargs)

        # Classify all explicitly specified parameters.
        self._signal = signal

    # ..................{ EMITTERS                           }..................
    def emit(self, record) -> None:

        # Log messaged formatted from this log record via this handler's current
        # log message formatter.
        record_message = self.format(record)

        # Redirect this log message to all slots connected to this handler's
        # previously initialized signal.
        self._signal.emit(record_message)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Concrete subclasses defining this application's command line interface (CLI).
'''

#FIXME: Display a splash screen in the BetseeCLI.__init__() method. Fortunately,
#Qt 5 makes this absurdly simple via the stock "QSplashScreen" class -- which
#comes self-provisioned with multi-threading capability, preventing us from
#having to interface with multi-threading merely to display a splash screen --
#which is pretty awesome, actually.
#
#It's trivially simple to use. The "QSplashScreen" documentation actually turns
#out to be the most useful help here. In particular, the workflow appears to
#resemble:
#
#* Define a new "betsee.util.app.guiappsplash" submodule containing:
#
#    from PySide2.QtCore import QPixmap
#    from PySide2.QtWidgets import QSplashScreen
#    from betsee.util.app.guiapp import APP_GUI
#
#    GUI_APP_SPLASH = None
#    '''
#    Singleton :class:`QSplashScreen`-based widget
#    '''
#
#    class QBetseeSplashScreen(QSplashScreen):
#        '''
#        :class:`QSplashScreen`-based widget displaying a non-modal splash
#        screen, typically used to present a multi-threaded loading screen
#        during time-consuming application startup.
#        '''
#
#        @type_check
#        def __init__(self, image_uri: str) -> None:
#
#            pixmap = QPixmap(image_uri)
#            super().__init__(pixmap)
#            self.show()
#            APP_GUI.processEvents()
#
#        def set_info(self, info: str) -> None:
#            '''
#            Display the passed human-readable single-line string as the
#            current progress message for this splash screen.
#            '''
#
#            self.showMessage(info)
#            APP_GUI.processEvents()
#
#
#* In this submodule, improve the existing
#
#    from betsee.util.widget import guisplash
#    from betsee.util.widget.guisplash import QBetseeSplashScreen
#
#
#    def __init__(self) -> None:
#
#        #FIXME: Non-ideal way to set a singleton, but... meh.
#        guisplash.GUI_SPLASH = QBetseeSplashScreen(
#            image_uri=':/GUI_SPLASH.png')
#
#        # Loading some items
#        guisplash.GUI_SPLASH.set_info("Loaded modules")
#
#        # Establishing connections
#        guisplash.GUI_SPLASH.set_info("Established connections")
#
#        #FIXME: This obviously needs to happen elsewhere. Since "GUI_SPLASH" is
#        #a globally accessible singleton, this shouldn't be an issue. *shrug*
#        # Create the main window.
#        main_window = QMainWindow()
#        guisplash.GUI_SPLASH.finish(main_window)
#
#        # Populate and show the main window.
#        main_window.show()
#        APP_GUI.exec_()
#        guisplash.GUI_SPLASH = None
#
#It's all pretty trivial stuff, frankly. Awesomeness!

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and application modules, including both
#   BETSEE and BETSE modules.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.util.cli.cliabc import CLIABC
from betse.util.cli.cliopt import CLIOptionArgStr
from betse.util.io.log import logs
from betse.util.type.types import type_check, ModuleType, SequenceTypes

# ....................{ SUBCLASS                          }....................
class BetseeCLI(CLIABC):
    '''
    Command line interface (CLI) for this application.

    Attributes
    ----------
    _sim_conf_filename : StrOrNoneTypes
        Absolute or relative filename of the initial YAML-formatted simulation
        configuration file to be initially opened by this application's GUI if
        any *or* ``None`` otherwise. This filename is parsed from command-line
        options passed by the current user.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self):

        # Initialize our superclass.
        super().__init__()

        # Nullify all instance variables for safety.
        self._sim_conf_filename = None

    # ..................{ SUPERCLASS ~ property : optional  }..................
    # The following properties *MAY* be implemented by subclasses.

    @property
    def _is_option_matplotlib_backend(self) -> bool:
        '''
        ``False``, preventing this CLI from exposing the
        ``--matplotlib-backend`` option and hence permitting users to
        externally specify an arbitrary matplotlib backend at the command line.

        Of necessity, this Qt-based application strictly requires a single
        Qt-based matplotlib backend (e.g., ``Qt5Agg``).
        '''

        return False

    # ..................{ SUPERCLASS ~ property : mandatory }..................
    # The following properties *MUST* be implemented by subclasses.

    @property
    def _module_ignition(self) -> ModuleType:

        # Defer heavyweight imports.
        from betsee import guiignition
        return guiignition


    @property
    def _module_metadata(self) -> ModuleType:

        # Defer heavyweight imports.
        from betsee import guimetadata
        return guimetadata


    @property
    def _help_epilog(self) -> str:

        return '''
To simulate any simulation produced by this GUI at the command line, consider
running the "betse" command underlying this GUI instead. For example, to
seed, initialize, and then simulate such a simulation in the current directory:

;    betse seed sim_config.yaml
;    betse init sim_config.yaml
;    betse  sim sim_config.yaml
'''

    # ..................{ SUPERCLASS ~ igniters             }..................
    def _init_app_libs(self) -> None:

        # Defer heavyweight imports.
        from betsee.lib import guilib
        from betsee.util.thread import guithread

        # Initialize all mandatory runtime dependencies of this application,
        # including both BETSE and BETSEE.
        #
        # Note that the superclass _init_app_libs() method is intentionally
        # *NOT* called, as that method sets the matplotlib backend. While doing
        # so is typically desirable, a chicken-and-egg conflict between Qt and
        # matplotlib complicates the initialization of both.
        #
        # See the body of the function called here for further details.
        guilib.reinit()

        # Initialize multithreading facilities *AFTER* validating dependencies,
        # which take precedence by virtue of their being things we depend upon.
        guithread.init()

    # ..................{ SUPERCLASS ~ options              }..................
    def _make_options_top(self) -> SequenceTypes:

        # Sequence of all default top-level options.
        options_top = super()._make_options_top()

        # Return a list extending this sequence with subclass-specific options.
        return list(options_top) + [
            CLIOptionArgStr(
                long_name='--sim-conf-file',
                synopsis='simulation configuration file to initially open',
                var_name='sim_conf_filename',
                default_value=None,
            ),
        ]


    def _parse_options_top(self) -> None:

        # Parse all default top-level options.
        super()._parse_options_top()

        # Initial simulation configuration file parsed from the passed options.
        self._sim_conf_filename = self._args.sim_conf_filename

    # ..................{ SUPERCLASS ~ methods              }..................
    def _do(self) -> object:
        '''
        Implement this command-line interface (CLI) by unconditionally running
        this application's graphical user interface (GUI), returning this
        interface to be memory profiled when the ``--profile-type=size`` CLI
        option is passed.
        '''

        # Defer imports *NOT* guaranteed to exist at this module's top-level.
        from betsee.gui.guimain import BetseeGUI

        # Application GUI.
        #
        # For safety, this GUI is scoped to a local rather than instance or
        # global variable, ensuring this GUI is implicitly destroyed by Python
        # before the root Qt application widget containing this GUI.
        app_gui = BetseeGUI(sim_conf_filename=self._sim_conf_filename)

        # Run this GUI's event loop and display this GUI, propagating the
        # returned exit status as this application's exit status.
        self._exit_status = app_gui.run()

        # Return this GUI for optional profiling purposes.
        return app_gui


    @type_check
    def _handle_exception(self, exception: Exception) -> None:

        # Defer to superclass handling, which typically logs this exception.
        super()._handle_exception(exception)

        # Additionally attempt to...
        try:
            # Import PySide2.
            from betsee.util.io import guierr

            # Display a PySide2-based message box displaying this exception.
            guierr.show_exception(exception)
        # If PySide2 or any other module indirectly imported above is
        # unimportable, print this exception message but otherwise ignore this
        # exception. Why? Because we have more significant fish to fry.
        except ImportError as import_error:
            logs.log_error(str(import_error))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level caching functionality for this application's graphical user
interface (GUI), persisting external resources required by this GUI to
user-specific files on the local filesystem.
'''

#FIXME: Still insufficient. Why? Because we need to automatically invalidate
#caches whenever any file in the BETSEE codebase changes. *sigh*

# ....................{ IMPORTS                           }....................
import PySide2
from betse.util.io.log import logs
from betse.util.path import files, paths, pathnames
from betse.util.path.command import cmds, cmdpath
from betse.util.py import pymodule, pys
from betse.util.type.types import type_check, IterableTypes
from betsee import guipathtree
from betsee.gui.simconf.stack.widget.guisimconfradiobtn import (
    QBetseeSimConfEnumRadioButtonGroup)
from betsee.lib import guilib
from betsee.util.io.xml import guiqrc, guiui

# ....................{ GLOBALS                           }....................
#FIXME: When upstream permits "QButtonGroup" widgets to be promoted via
#Qt (Creator|Designer), remove this ad-hack kludge.
_PROMOTE_OBJ_NAME_TO_CLASS = {
    # Manually promoto "QButtonGroup" widgets to application-specific types.
    'sim_conf_space_intra_lattice_type': QBetseeSimConfEnumRadioButtonGroup,
}
'''
Dictionary mapping from the name of each instance variable of the main window
to the application-specific widget subclass to declare that variable to be an
instance of.

This dictionary facilitates the manual "promotion" of widgets for which the
Qt (Creator|Designer) GUI currently provides no means of official promotion,
notably including :class:`QButtonGroup` widgets.
'''

# ....................{ CACHERS                           }....................
def cache_py_files() -> None:
    '''
    Either create and cache *or* reuse each previously cached pure-Python
    module required at runtime by this GUI, including all :mod:`PySide2`-based
    modules converted from XML-formatted files and binary resources exported by
    the external Qt Designer GUI.

    For efficiency, previously generated modules are regenerated *only* as
    needed (i.e., if older than the underlying XML files and other associated
    paths from which these modules are generated).
    '''

    # Append the directory containing all generated modules to the PYTHONPATH,
    # permitting these modules to be subsequently imported elsewhere.
    pys.add_import_dirname(guipathtree.get_data_py_dirname())

    # Generate the requisite pure-Python modules (in any arbitrary order).
    _cache_py_qrc_file()
    _cache_py_ui_file()

    # For safety, raise an exception unless all such modules exist now.
    files.die_unless_file(
        guipathtree.get_data_py_qrc_filename(),
        guipathtree.get_data_py_ui_filename())

# ....................{ CACHERS ~ private                 }....................
def _cache_py_qrc_file() -> None:
    '''
    Reuse the previously cached pure-Python :mod:`PySide2`-based module
    embedding all binary resources in this application's main Qt resource
    collection (QRC) if this module is sufficiently up-to-date (i.e., at least
    as new as all input paths required to regenerate this module) *or*
    regenerate this module from these input paths otherwise.

    Raises
    ----------
    BetseFileException
        If this module is outdated (i.e., older than at least one input path
        required to regenerate this module) but is unwritable by the current
        user, in which case this module is *NOT* updateable and is thus
        desynchronized from the remainder of the codebase. Because this
        desynchronization is liable to induce subtle non-human-readable issues,
        a fatal exception is raised rather than a non-fatal warning logged.

    See Also
    ----------
    :func:`_is_output_path_outdated`
        Further details.
    '''

    # Absolute path of the input QRC file used to generate this output module.
    data_qrc_filename = guipathtree.get_data_qrc_filename()

    # Absolute path of the output module to be generated.
    data_py_qrc_filename = guipathtree.get_data_py_qrc_filename()

    # List of the absolute pathnames of all input paths required to do so. For
    # efficiency, these paths are ordered according to the heuristic discussed
    # by the paths.is_mtime_recursive_older_than_paths() function.
    input_pathnames = [
        data_qrc_filename,
    ]

    # If the optional third-party dependency "pyside2-tools" is installed,
    # append the "pyside2-rcc" executable for testing as well.
    if cmds.is_command('pyside2-rcc'):
        input_pathnames.append(cmdpath.get_filename('pyside2-rcc'))

    # If this output module is at least as new as *ALL* the following paths,
    # this output module is sufficiently up-to-date and need *NOT* be
    # regenerated:
    #
    # * The input "pyside2-rcc" executable run by the
    #   psdqrc.convert_qrc_to_py_file() function called below.
    # * Any file or subdirectory in the input directory containing both this
    #   input QRC file and all resource files referenced by this file.
    if not _is_output_path_outdated(
        input_pathnames=input_pathnames, output_filename=data_py_qrc_filename):
        return

    # Else, this output module is older than at least one such path, in which
    # case this output module is outdated and must be regenerated.
    guiqrc.convert_qrc_to_py_file_if_able(
        qrc_filename=data_qrc_filename, py_filename=data_py_qrc_filename)


def _cache_py_ui_file() -> None:
    '''
    Reuse the previously cached pure-Python :mod:`PySide2`-based module
    implementing the superficial construction of this application's main window
    if this module is sufficiently up-to-date (i.e., at least as new as all
    input paths required to regenerate this module) *or* regenerate this module
    from these input paths otherwise.

    Raises
    ----------
    BetseFileException
        If this module is outdated (i.e., older than at least one input path
        required to regenerate this module) but is unwritable by the current
        user, in which case this module is *NOT* updateable and is thus
        desynchronized from the remainder of the codebase. Because this
        desynchronization is liable to induce subtle non-human-readable issues,
        a fatal exception is raised rather than a non-fatal warning logged.

    See Also
    ----------
    :func:`_is_output_path_outdated`
        Further details.
    '''

    # Absolute path of the input QRC file used to generate this output module.
    data_ui_filename = guipathtree.get_data_ui_filename()

    # Absolute path of the output module to be generated.
    data_py_ui_filename = guipathtree.get_data_py_ui_filename()

    # List of the absolute pathnames of all input paths required to do so. For
    # efficiency, these paths are ordered according to the heuristic discussed
    # by the paths.is_mtime_recursive_older_than_paths() function.
    input_pathnames = [
        data_ui_filename,
        pymodule.get_filename(guiui),
        pymodule.get_dirname(PySide2),
    ]

    # If the optional third-party dependency "pyside2-tools" is installed...
    if guilib.is_runtime_optional('pyside2uic'):
        # Package installed by this dependency.
        pyside2uic = guilib.import_runtime_optional('pyside2uic')

        # Append this package's directory for testing as well.
        input_pathnames.append(pymodule.get_dirname(pyside2uic))

    # If this output module is at least as new as *ALL* the following paths,
    # this output module is sufficiently up-to-date and need *NOT* be
    # regenerated:
    #
    # * This input UI file.
    # * The file providing the submodule of this application converting this UI
    #   file into a Python module.
    # * Any file or subdirectory in the input directories containing the
    #   "PySide2" and "pyside2uic" packages required by the
    #   psdui.convert_ui_to_py_file() function called below.
    if not _is_output_path_outdated(
        input_pathnames=input_pathnames, output_filename=data_py_ui_filename):
        return

    # Else, this output module is older than at least one such path, in which
    # case this output module is outdated and must be regenerated.
    guiui.convert_ui_to_py_file_if_able(
        ui_filename=data_ui_filename,
        py_filename=data_py_ui_filename,
        promote_obj_name_to_class=_PROMOTE_OBJ_NAME_TO_CLASS,
    )

# ....................{ TESTERS ~ private                 }....................
@type_check
def _is_output_path_outdated(
    input_pathnames: IterableTypes, output_filename: str) -> bool:
    '''
    ``True`` only if the output path either does not exist, does but is older
    than all input paths in the passed iterable, *or* does but is an empty
    (i.e., zero-byte) file.

    If this function returns ``True``, the caller is expected to explicitly
    (re)create this output path from these input paths.

    Parameters
    ----------
    input_pathnames: IterableTypes[str]
        Iterable of the absolute or relative pathnames of all input paths
        required to (re)create this output path. For efficiency, these paths
        should be ordered according to the heuristic discussed by the
        :func:`paths.is_mtime_recursive_older_than_paths` function.
    output_filename : str
        Absolute or relative pathname of the output path.
    '''

    # Log this inspection.
    logs.log_info(
        'Synchronizing cached submodule "%s"...',
        pathnames.get_basename(output_filename))

    # Return true only if either...
    return (
        # This output module does not exist *OR*...
        not paths.is_path(output_filename) or
        # This output module does exist but is older than at least one of these
        # output paths *OR*...
        paths.is_mtime_recursive_older_than_paths(
            output_filename, input_pathnames) or
        # This output module is a zero-byte file.
        files.get_size(output_filename) == 0
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Submodule both instantiating and initializing the :class:`QApplication`
singleton for this application.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and BETSEE modules. This does *NOT*
#   include BETSE modules, which are *NOT* guaranteed to exist at this point.
#   For simplicity, however, all core PySide2 submodules are assumed to exist.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import logging
from PySide2.QtCore import Qt, QCoreApplication
from PySide2.QtGui import QGuiApplication
from PySide2.QtWidgets import QApplication, qApp
from betsee import guimetadata
from betsee.guiexception import BetseePySideApplicationException

# ....................{ GLOBALS                           }....................
_IS_INITTED = False
'''
``True`` only if the :func:`init` function has already been called.

That function uses this private boolean to guard against repeated invocations
of the :func:`init` function from multiple modules in the same Python process.
Unlike the comparable :func:`betse.lib.libs.init` function, the :func:`init`
function defined by this submodule does *not* support repeated calls and hence
*must* explicitly guard against that uncommon edge case.
'''

# ....................{ EXCEPTIONS                        }....................
def die_unless_initted() -> None:
    '''
    Raise an exception unless the :func:`init` function has been called and
    hence instantiated the :class:`QApplication` singleton for this
    application.

    For safety, this function additionally raises an exception if the
    :func:`init` function has been called but the :class:`QApplication`
    singleton instantiated by that function no longer exists.

    Raises
    ----------
    BetseePySideApplicationException
        If either:
        * The :func:`init` function has yet to be called.
        * The :class:`QApplication` singleton has yet to be instantiated.
    '''

    # If this function has yet to be be called, raise an exception.
    if not _IS_INITTED:
        raise BetseePySideApplicationException(QCoreApplication.translate(
            'guiapp',
            'betsee.util.app.guiapp.init() has yet to be called.'))

    # If the "QApplication" singleton is uninstantiated. raise an exception.
    if qApp is None:
        raise BetseePySideApplicationException(QCoreApplication.translate(
            'guiapp',
            '"QApplication" singleton uninstantiated '
            '(e.g., as betsee.util.app.guiapp.init() has yet to be called).'))

# ....................{ GETTERS                           }....................
def get_app() -> QApplication:
    '''
    Initialized :class:`QApplication` singleton for this application if
    instantiated by a prior call to the :func:`init` function *or* raise an
    exception otherwise (i.e., if that function has yet to be called).

    Contrary to nomenclature, the :class:`QApplication` class confusingly
    subclasses the :class:`QGuiApplication` base class in a manner optimized
    for widgets. Ergo, the former is *always* preferable to the latter.

    Design
    ----------
    This function does *not* implicitly call the :func:`init` function if the
    :class:`QApplication` singleton has yet to be instantiated. While
    non-intuitive and arguably inconvenient, this is intentional. Implicitly
    calling that function would technically be trivial but invite subtle (and
    hence non-debuggable) issues; in particular, the order in which the
    :func:`init` and :func:`betse.lib.libs.reinit` methods are called is very
    significant and hence must *not* be left up to non-deterministic chance.

    Caveats (Exceptions)
    ----------
    Avoid directly accessing the low-level :attr:`PySide2.QtWidgets.qApp`
    global, which this higher-level function wraps. Since that global is
    typically ``None`` prior to the first call to the :func:`init` function,
    directly accessing that global sufficiently early in runtime may raise
    non-human-readable exceptions *or* induce low-level segmentation faults
    from Qt itself resembling:

        QWidget: Must construct a QApplication before a QWidget
        [1]    30475 abort      betsee -v

    In contrast, this function *always* raises human-readable exceptions.

    Caveats (Attributes)
    ----------
    Avoid attempting to add application-specific instance variables to the
    low-level :attr:`PySide2.QtWidgets.qApp` global returned by this function,
    as that global silently ignores all such attempts. While this constraint
    could technically be circumvented by globally persisting the
    :class:`QApplication` singleton created by the :func:`init` function as a
    module-scoped Python singleton, doing so would incur subtle issues of its
    own -- including complications in both garbage collection and accidental
    collision with standard Qt attributes.

    Returns
    ----------
    QApplication
        Initialized :class:`QApplication` singleton for this application.

    Raises
    ----------
    BetseePySideApplicationException
        If either:
        * The :func:`init` function has yet to be called.
        * The :class:`QApplication` singleton has yet to be instantiated.
    '''

    # If either the init() funtion has yet to be called *OR* the "QApplication"
    # singleton is uninstantiated. raise an exception.
    die_unless_initted()

    # Else, this singleton is *NOT* "None". Return this singleton.
    return qApp

# ....................{ INITIALIZERS                      }....................
def init() -> None:
    '''
    Instantiate and initialize the :class:`QApplication` singleton for this
    application (i.e., the :attr:`PySide2.QApplication.qApp` instance) if this
    function has not already been called *or* silently reduce to a noop
    otherwise (i.e., if this function has already been called).
    '''

    # If this function has already been called, noop.
    global _IS_INITTED
    if     _IS_INITTED:
        return

    # Log this initialization.
    logging.debug('Instantiating Qt application singleton...')

    # Destroy the existing "QApplication" singleton (if any).
    _deinit_qt_app()

    # Set static attributes of the "QApplication" class *BEFORE* defining the
    # singleton instance of this class.
    _init_qt()

    # Instantiate a singleton of this class.
    _init_qt_app()

    # Record this function as having been called *AFTER* successfully doing so.
    _IS_INITTED = True

# ....................{ DEINITIALIZERS                     }....................
#FIXME: Actually, this error appears to be induced by initializing third-party
#BETSE libraries and hence the "Qt5Agg" backend *BEFORE* instantiating this
#singleton here. Clearly, the order of these two operations needs to be
#reversed. After that is done, this function should be:
#
#* Renamed to die_if_qt_app().
#* Refactored to raise a human-readable exception, which should explicitly note
#  the likelihood of a previously imported Python package (e.g., "matplotlib")
#  having externally instantiated this singleton already.
#* Redocumented as such.

def _deinit_qt_app() -> None:
    '''
    Destroy the existing :class:`QApplication` singleton with a non-fatal
    warning if such a singleton has been previously initialized elsewhere *or*
    silently reduce to a noop otherwise.

    While this condition should arguably constitute a fatal error inducing a
    raised exception, various versions of PySide2 appear to erroneously
    initialize this singleton on first importation without our explicit
    consent. There isn't much we can do about it; this is the next best thing.

    If this singleton is _not_ explicitly destroyed, PySide2 raises the
    following exception on attempting to re-initialize another such singleton:

        RuntimeError: Please destroy the QApplication singleton before creating
        a new QApplication instance.
    '''

    # Existing "QApplication" singleton if any or "None" otherwise.
    app_prior = QCoreApplication.instance()

    # If an existing "QApplication" singleton has already been initialized...
    if app_prior is not None:
        # Log a non-fatal warning.
        logging.warning(
            'Destroying erroneously instantiated Qt application singleton...')

        # Destroy this singleton.
        app_prior.quit()

# ....................{ INITIALIZERS : qt                 }....................
def _init_qt() -> None:
    '''
    Initialize static attributes of the :class:`QApplication` class or
    subclasses thereof (e.g., :class:`QCoreApplication`,
    :class:`QGuiApplication`) *before* the singleton instance of this class is
    defined.

    Technically, some of these attributes (e.g.,
    :attr:`Qt.AA_UseHighDpiPixmaps`) are safely definable at any time. Since
    others (e.g. ,:attr:`Qt.AA_EnableHighDpiScaling`) are *not*, all such
    attributes are preemptively defined here for both simplicity and safety.

    These attributes pertain to the :class:`QApplication` singleton rather than
    this singleton's :class:`QMainWindow` instance implemented by the
    XML-formatted UI file exported by Qt Creator; thus, these attributes
    *cannot* be specified by this file but *must* instead be manually
    implemented in Python.
    '''

    # Avoid circular import dependencies.
    from betsee.util.io import guisettings

    # Initialize all application-wide core attributes (e.g., name, version).
    _init_qt_metadata()

    # Initialize all application-wide dots per inch (DPI) attributes.
    _init_qt_dpi()

    # Initialize all application-wide QSettings attributes.
    guisettings.init()


def _init_qt_metadata() -> None:
    '''
    Initialize all static attributes of the :class:`QCoreApplication` class
    signifying application-wide core properties (e.g., name, version).
    '''

    # Log this initialization.
    logging.debug('Initializing static Qt attributes...')

    # High-level human-readable application name intended *ONLY* for display.
    QGuiApplication.setApplicationDisplayName(guimetadata.NAME)

    # Low-level machine-readable application name and version, each intended
    # both for display (e.g., to end users) and internal inspection.
    QCoreApplication.setApplicationName(guimetadata.NAME)
    QCoreApplication.setApplicationVersion(guimetadata.VERSION)

    # Low-level machine-readable organization name and domain.
    QCoreApplication.setOrganizationName(guimetadata.ORG_NAME)
    QCoreApplication.setOrganizationDomain(guimetadata.ORG_DOMAIN_NAME)


def _init_qt_dpi() -> None:
    '''
    Initialize all static attributes of the :class:`QApplication` class
    pertaining to dots per inch (DPI) and, specifically, high-DPI displays.

    See Also
    ----------
    https://blog.qt.io/blog/2016/01/26/high-dpi-support-in-qt-5-6
        *High-DPI Support in Qt 5.6,* article colloquially describing the
        attributes initialized by this method.
    '''

    # Attempt to...
    try:
        # Import BETSE submodules, whose importability has yet to be validated.
        from betse.util.os import oses
        from betse.util.os import displays

        # If none of the following conditions is satisfied:
        #
        # * The current platform is macOS, which natively supports high-DPI
        #   scaling out-of-the-box. Moreover, the official documentation for
        #   the Qt attribute set below explicitly states: "Supported platforms
        #   are X11, Windows and Android." For safety, this attribute is *NOT*
        #   enabled under macOS.
        # * The current platform is Linux and the current display server is a
        #   Wayland compositor, which all natively support high-DPI scaling.
        #
        # Then the current display environment does *NOT* natively support
        # high-DPI scaling. Notably, the Windows and X11 display environments
        # both fail to do so. In this case, we inform Qt that it should attempt
        # to do so via emulation at the framework level, converting all
        # previously physical pixels defined throughout this application into
        # logical pixels portable across displays sporting varying DPI.
        if not (oses.is_macos() or displays.is_linux_wayland()):
            logging.debug('Initializing high-DPI scaling emulation...')
            QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    # If any such submodule is unimportable, log a non-fatal error and
    # continue.  Since BETSE is a mandatory dependency, its unimportability
    # would typically constitute a fatal error. Since subsequent dependency
    # checking is guaranteed to raise a human-readable exception on our behalf,
    # however, we needn't uselessly duplicate this checking here.
    except ImportError as exception:
        logging.error(str(exception))

    # Permit the QIcon.pixmap() method to generate high-DPI pixmaps larger
    # than the requested size (i.e., a devicePixelRatio() larger than 1.0).
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)

# ....................{ INITIALIZERS : globals            }....................
def _init_qt_app() -> QApplication:
    '''
    Instantiate and return the :class:`QApplication` singleton for this
    application.
    '''

    # Avoid circular import dependencies.
    # from betsee.util.filter.guifiltertooltip import (
    #     QBetseePlaintextTooltipEventFilter)

    # Log this instantiation.
    logging.debug('Instantiating application singleton...')

    #FIXME: Actually, this isn't *QUITE* right. At least the "-session" option
    #should be forwarded on to the "QApplication" constructor, required for
    #application restoration after having been previously suspended. See the
    #QSessionManager::setRestartCommand() method documentation for details.
    #
    #Note that this option *MUST* be named "-session" unless the
    #setRestartCommand() method is called to override this default. We'd
    #certainly prefer a *nix-style option named "--session-id" instead,
    #however. Can we make this happen?

    # For safety, initialize this application with *NO* command-line arguments
    # regardless of whether the current CLI was passed arguments. The subset of
    # arguments parsed by this widget are of no interest to end users and
    # developers alike. Since this object consumes all arguments it parses,
    # permitting this object to arbitrarily parse and hence consume arguments
    # encourages conflicts with future versions of Qt. In theory, Qt could
    # expand the subset of arguments parsed by this object to those already
    # parsed by the current CLI! That's bad.
    gui_app = QApplication([])

    #FIXME: Document why this might be an awful idea: notably, the GIL blocking
    #multithreaded event handling. If this is indeed one of several culprits
    #(which seems fairly likely), we'll want to revise our original
    #StackOverflow answer advising usage of this paradigm under Python. Clearly,
    #this approach only applies to languages other than Python -- notably, C++.
    #FIXME: O.K.; we've verified by manual inspection that this event filter
    #does indeed slow event handling down by approximately 200% when attempting
    #to perform multithreading -- clearly due to the GIL. Given that, we'll need
    #to manually implement the equivalent of this filter by:
    #
    #* For each plaintext (i.e., non-HTML) tooltip across the entire
    #  application...
    #  * Coercively embed that tooltip in HTML tags.
    #
    #To do so sanely, we'll probably want to perform a simple global
    #search-and-replace in our "betsee.ui" file. *sigh*

    # Install an application-wide event filter globally addressing severe
    # issues in Qt's default plaintext tooltip behaviour.
    # gui_app.installEventFilter(QBetseePlaintextTooltipEventFilter(gui_app))

    # Return this application.
    return gui_app



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Top-level classes defining this application's main window.

Caveats
----------
This submodule imports from the module whose fully-qualified name is given by
:attr:`metadata.MAIN_WINDOW_UI_MODULE_NAME`, an external top-level module
dynamically generated at runtime from XML-formatted files exported by the
external Qt Designer GUI and bundled with this application's codebase. Since
that module is *not* guaranteed to exist at application startup, this
submodule is safely importable only *after* the :mod:`betsee.gui.guicache`
submodule has locally created and cached that module for the current user.
'''

#FIXME: Consider permitting multiple simulations to be simultaneously open. The
#conventional means of doing so is the so-called Single Documentation Interface
#(SDI) approach, in which opening a new document opens a new distinct
#application window displaying that new document; open documents and hence
#windows may then be switched between via a "Windows" menu item listing all
#currently open documents. Supporting SDI would permit end users to run multiple
#simultaneous simulations, which only seems reasonable. As a PySide2-specific
#example, see the "pyside2-examples/examples/widgets/mainwindows/sdi/sdi.py"
#application. While incomplete, this application should serve as a simple
#starting point for implementing our own SDI. (It's only 295 lines!)
#FIXME: Actually, modern UI design strongly favours "QTabWidget"-style
#browser-based multiplicity -- and so do we. Ergo, classical window-style
#SDI- and MDI-based multiplicity are right out. That said, it shouldn't be
#terribly arduous to add a new top-level "QTabWidget" permitting multiple
#concurrently open simulations to be switched between. (Yeah, right.)

#FIXME: The status bar of the main window should be dynamically modified
#whenever *ANY* action (e.g., simulation opening or closure) is successfully
#completed. See to it, please.

#FIXME: The title of the main window should probably be dynamically modified on
#opening and closing a simulation configuration file to reflect both the
#basename of this file *AND* a character indicating whether this file has
#unsaved changes or not. Courtesy the official documentation for the
#"QWidget.windowModified" boolean property:
#
#    "The window title must contain a "[*]" placeholder, which indicates where
#     the '*' should appear. Normally, it should appear right after the file
#     name (e.g., "document1.txt[*] - Text Editor"). If the window isn't
#     modified, the placeholder is simply removed.
#
#     On some desktop platforms (including Windows and Unix), the application
#     name (from QGuiApplication::applicationDisplayName) is added at the end of
#     the window title, if set. This is done by the QPA plugin, so it is shown
#     to the user, but isn't part of the windowTitle string."

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific submodules of this subpackage (i.e.,
# "betsee.gui.window"). Since those submodules must *ALWAYS* be able to safely
# import from this submodule, circularities are best avoided here.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# Accessing attributes of the "PySide2.QtCore.Qt" subpackage requires this
# subpackage to be imported directly. Attributes are *NOT* importable from this
# subpackage, contrary to pure-Python expectation.
from PySide2.QtCore import Qt, Slot
from PySide2.QtGui import QCloseEvent
from PySide2.QtWidgets import QToolButton, QWidget
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.obj import objects
from betse.util.type.types import type_check, StrOrNoneTypes
from betsee import guimetadata
from betsee.guiexception import BetseePySideWindowException
from betsee.gui.guisignal import QBetseeSignaler
from betsee.util.app import guiappwindow
from betsee.util.io import guierr
from betsee.util.io.log import guilogconf
from betsee.util.io.xml import guiui
from betsee.util.type.guitype import QWidgetOrNoneTypes

# ....................{ GLOBALS                           }....................
MAIN_WINDOW_BASE_CLASSES = guiui.get_ui_module_base_classes(
    ui_module_name=guimetadata.MAIN_WINDOW_UI_MODULE_NAME)
'''
Sequence of all main window base classes declared by the module whose
fully-qualified name is given by :attr:`metadata.MAIN_WINDOW_UI_MODULE_NAME`.
'''

# ....................{ CLASSES                           }....................
# Subclass all main window base classes declared by the above module (in
# order).  While multiple inheritance typically invites complex complications
# (e.g., diamond inheritance problem) and hence is best discouraged, these base
# classes are guaranteed *NOT* to conflict in this manner. All alternatives to
# this multiple inheritance design invite worse complications and otherwise
# avoidable annoyances. In short, this is arguably the best we can do.
#
# For further details, see the following classic PyQt treatise:
#     http://pyqt.sourceforge.net/Docs/PyQt5/designer.html
class QBetseeMainWindow(*MAIN_WINDOW_BASE_CLASSES):
    '''
    Main window Qt widget for this application, doubling as both this
    application's root Qt widget containing all other Qt widgets.

    Design
    ----------
    This class subclasses all main window base classes declared by the module
    whose fully-qualified name is given by
    :attr:`metadata.MAIN_WINDOW_UI_MODULE_NAME`. While multiple inheritance
    often invites complex complications (e.g., diamond inheritance) and hence
    is best discouraged, these base classes are guaranteed *not* to conflict in
    this manner. All alternatives to this multiple inheritance design invite
    worse complications and otherwise avoidable annoyances. In short, this is
    arguably the best we can do.

    Attributes (Public)
    ----------
    signaler : QBetseeSignaler
        :class:`PySide2`-based collection of various application-wide signals.
    sim_conf : QBetseeSimConf
        Object encapsulating high-level simulation configuration state.
    sim_tab : QBetseeSimmerTabWidget
        Object encapsulating high-level simulator state.

    Attributes (Private)
    ----------
    _clipboard : QBetseeMainClipboard
        Object encapsulating high-level application clipboard state.
    _sim_conf_filename : StrOrNoneTypes
        Absolute or relative path of the initial YAML-formatted simulation
        configuration file to be initially opened if any *or* ``None``
        otherwise.

    See Also
    ----------
    http://pyqt.sourceforge.net/Docs/PyQt5/designer.html
        :mod:`PyQt5`-specific documentation detailing this design.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        signaler: QBetseeSignaler,
        sim_conf_filename: StrOrNoneTypes,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this main window.

        Parameters
        ----------
        signaler : QBetseeSettingsSignaler
            :class:`PySide2`-based collection of application-wide signals.
        sim_conf_filename : StrOrNoneTypes
            Absolute or relative path of the initial YAML-formatted simulation
            configuration file to be initially opened if any *or* ``None``
            otherwise.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Classify all remaining parameters.
        self.signaler = signaler
        self._sim_conf_filename = sim_conf_filename

        # Nullify all remaining instance variables for safety.
        #
        # Note that the "sim_tab" instance variable is defined by our
        # superclass, itself statically defined by our main UI file and hence
        # "betsee.data.py.betsee_ui" submodule. For obvious reasons, this
        # instance variable must *NOT* be overridden to "None" here.
        self.sim_conf = None
        self._clipboard = None

        # Log this initialization.
        logs.log_debug('Generating main window...')

        # Customize this main window as specified by the XML-formatted UI file
        # exported by Qt Creator. This superclass method is defined by the
        # helper base class generated by the "betsee.lib.pyside.psdui" module.
        self.setupUi(self)

        # Customize this main window with additional Python logic.
        self._init()


    def _init(self) -> None:
        '''
        Customize this main window with additional Python logic *after*
        customizing this main window as specified by the XML-formatted UI file
        exported by Qt Creator.

        Design
        ----------
        **Customizations implemented by this method must not be implementable
        by this UI file.** Equivalently, this method should *only* implement
        customizations requiring dynamic Python logic, which typically implies
        signals and slots. For portability, all other customizations should be
        implemented by this UI file.
        '''

        # Append all unfiltered log records to the top-level log widget in an
        # autoscrolling, non-blocking, thread-safe manner *BEFORE* performing
        # any subsequent logic possibly performing logging.
        guilogconf.log_to_text_edit(self.log_box)

        # Customize all abstract actions of this main window.
        self._init_actions()

        # Customize all physical top-level widgets of this main window *AFTER*
        # customizing all abstract actions required by these widgets.
        self._init_widgets()

        # Finalize the initialization of this main window *AFTER* customizing
        # all physical and abstract widgets of this window .
        self._init_end()

    # ..................{ INITIALIZERS ~ actions            }..................
    def _init_actions(self) -> None:
        '''
        Customize all abstract :class:`QAction` widgets of this main window,
        typically by associating the predefined C-based signals of these
        widgets with custom pure-Python slot callables.
        '''

        # Customize all actions in the "File", "Edit", and "Help" menus.
        self._init_actions_file()
        self._init_actions_edit()
        self._init_actions_help()

        #FIXME: Implement all remaining actions.

        # Default all remaining actions to display an error box informing the
        # end user that this action has yet to be implemented.
        for action in (
            # "Edit" menu.
            self.action_edit_prefs,

            # "Help" menu.
            self.action_about_betse,
            self.action_about_betsee,
        ):
            action.triggered.connect(self._show_error_action_unimplemented)


    def _init_actions_file(self) -> None:
        '''
        Customize all QAction widgets of the top-level ``File`` menu.
        '''

        # Associate action signals with corresponding slots.
        self.action_exit.triggered.connect(self.close)


    def _init_actions_edit(self) -> None:
        '''
        Customize all QAction widgets of the top-level ``Edit`` menu.
        '''

        pass


    def _init_actions_help(self) -> None:
        '''
        Customize all QAction widgets of the top-level ``Help`` menu.
        '''

        pass


    #FIXME: Excise this, which should no longer be required.
    def _show_error_action_unimplemented(self) -> None:
        '''
        Display a modal message box informing the end user that the currently
        selected action has yet to be implemented.
        '''

        guierr.show_error(
            title='Action Unimplemented',
            synopsis='This action is currently unimplemented.',
        )

    # ..................{ INITIALIZERS ~ widgets            }..................
    def _init_widgets(self) -> None:
        '''
        Customize all physical top-level widgets of this main window, thus
        excluding all abstract widgets (e.g., :class:`QAction`).
        '''

        # Avoid circular import dependencies.
        from betsee.gui.window.guimainclipboard import QBetseeMainClipboard
        from betsee.gui.simconf.guisimconf import QBetseeSimConf

        # Object wrapping high-level state, instantiated in arbitrary order.
        self._clipboard = QBetseeMainClipboard()
        self.sim_conf = QBetseeSimConf()

        # Initialize these objects (in arbitrary order) *BEFORE* initializing
        # widgets assuming these objects to have been initialized.
        self._clipboard.init(main_window=self)
        self.sim_conf.init(main_window=self)
        self.sim_tab.init(main_window=self)

        # Initialize the simulation configuration stack widget *BEFORE* its
        # higher-level sibling tree widget, which assumes the former to have
        # been initialized.
        self.sim_conf_stack.init(main_window=self)

        # Initialize the simulation configuration tree widget.
        self.sim_conf_tree.init(main_window=self)

    # ..................{ INITIALIZERS ~ end                }..................
    def _init_end(self) -> None:
        '''
        Finalize the initialization of this main window.

        As this method's name implies, this method is called *after* all
        widgets of this main window have been initialized. This method then
        performs all remaining logic assuming these widgets to be initialized
        -- including:

        * **Toolbar button sanitization.** Since widget initialization
          typically programmatically inserts additional buttons into the
          toolbar, sanitizing these buttons *must* be deferred until this
          method.
        '''

        # Prevent toolbar buttons from receiving the keyboard input focus,
        # ensuring that the currently focused widget if any retains focus even
        # when the user clicks one or more of these buttons. For safety, this
        # is done *AFTER* finalizing all widgets and hence programmatically
        # inserting all remaining toolbar buttons into the toolbar.
        #
        # By default, toolbar buttons are *ALWAYS* permitted to receive focus
        # regardless of whether the toolbar containing these buttons is
        # permitted to do so (i.e., has a focus policy of "NoFocus") and
        # despite there being no demonstrable use case for focusing these
        # buttons.
        for toolbar_button in self.toolbar.findChildren(QToolButton):
            toolbar_button.setFocusPolicy(Qt.NoFocus)

        #FIXME: For maintainability, refactor this as follows:
        #
        #* Remove the "self.set_filename_signal.emit('')" call from the
        #  "guisimconf" submodule.
        #* Define a new QBetseeSimConf.unload() method, called below. This
        #  method should internally perform the equivalent of:
        #    self.set_filename_signal.emit('')
        #* Expand the following if conditional to resemble:
        #    if self._sim_conf_filename is not None:
        #        self.sim_conf.load(self._sim_conf_filename)
        #    else:
        #        self.sim_conf.unload()

        # If opening an initial simulation configuration file, do so as this
        # method's last logic (i.e., *AFTER* all finalization performed above).
        if self._sim_conf_filename is not None:
            self.sim_conf.load(self._sim_conf_filename)

    # ..................{ TESTERS                           }..................
    def _is_closable(self) -> bool:
        '''
        ``True`` only if this application is **safely closable** (i.e.,
        closable without incuring possible simulation data loss).

        Depending on the current state of this application, this method may
        indefinitely block until the user interactively confirms all of the
        following safety constraints:

        * That all unsaved changes (if any) are to be saved for the currently
          open simulation (if any).
        * That the currently running simulation subcommand (if any) is to be
          prematurely halted.

        Design
        ----------
        This method is intentionally implemented as a method rather than
        property to imply to callers that this method incurs side effects,
        including indefinitely blocking until receiving user confirmation.
        '''

        # If this application has yet to be fully initialized, this application
        # is safely closable. Why? By definition, an application that is only
        # partially initialized is an application that has yet to be displayed
        # to the user and hence has yet to open and hence edit a simulation.
        if (
            objects.get_attr_or_none(obj=self, attr_name='sim_conf') is None or
            objects.get_attr_or_none(obj=self, attr_name='sim_tab') is None
        ):
            return True

        #FIXME: Also interactively confirm that the currently running
        #simulation subcommand (if any) is to be prematurely halted.

        # Return true only if the user interactively confirms...
        return (
            # That all unsaved changes (if any) are to be saved for the
            # currently open simulation (if any).
            self.sim_conf.save_if_dirty()
        )

    # ..................{ EVENTS                            }..................
    def closeEvent(self, event: QCloseEvent) -> None:
        '''
        Event handler handling the passed close event signifying a user-driven
        request to close this main window and exit the current application.

        See Also
        ----------
        :meth:`QObject.destroyed`
            Slot whose signals are signalled immediately *before* this object
            and all children objects of this object are destroyed.
        :meth:`QGuiApplication::lastWindowClosed`
            Slot whose signals are signalled immediately *before* the last main
            window (i.e., this singleton) and all children objects of this
            window are destroyed.
        '''

        # Log this attempt.
        logs.log_info('Attempting PySide2 UI closure...')

        # If this application is safely closable...
        if self._is_closable():
            # Log this closure.
            logs.log_info('Performing PySide2 UI closure...')

            # Halt all currently working simulator workers if any by attempting
            # to gracefully stop each such worker if feasible or non-gracefully
            # terminating each such worker otherwise.
            self.sim_tab.halt_work()

            # Store application-wide settings *BEFORE* closing this window.
            self.signaler.store_settings_signal.emit()

            # Unset the global exposing this main window, minimizing subtle
            # garbage collection issues during application destruction.
            guiappwindow.unset_main_window()

            # Accept this request, thus finalizing the closure of this window.
            # To ensure superclass handling is performed, call the superclass
            # implementation rather than event.accept().
            super().closeEvent(event)
        # Else, refuse this request, preventing this window from being closed.
        # Well, the window manager typically ignores us and closes the window
        # anyway... but hey. If it's battery life or us, we gotta go.
        else:
            # Log this refusal.
            logs.log_info('Ignoring PySide2 UI closure...')

            # Refuse this closure request.
            event.ignore()

    # ..................{ SLOTS ~ sim conf                  }..................
    @Slot(str)
    def set_sim_conf_filename(self, sim_conf_filename: StrOrNoneTypes) -> None:
        '''
        Slot invoked in response to both the opening of new simulation
        configurations *and* the closing of open simulation configurations.

        Parameters
        ----------
        sim_conf_filename : StrOrNoneTypes
            Absolute path of the currently open YAML-formatted simulation
            configuration file if any *or* ``None`` otherwise.
        '''

        # If a simulation configuration is currently open...
        if sim_conf_filename is not None:
            # Basename of this configuration file.
            sim_conf_basename = pathnames.get_basename(sim_conf_filename)

            # Update this window title to contain this basename appended by a
            # Qt-specific suffix to be subsequently replaced by a
            # platform-specific identifier when the setWindowModified() method
            # is called elsewhere.
            self.setWindowTitle('{}[*]'.format(sim_conf_basename))
        # Else, no simulation configuration is currently open. In this case,
        # clear this window title.
        else:
            self.setWindowTitle('')


    @Slot(bool)
    def set_sim_conf_dirty(self, is_sim_conf_dirty: bool) -> None:
        '''
        Slot invoked on the currently open simulation configuration associated
        with the main window either receiving new unsaved changes (in which
        case this boolean is ``True``) *or* having just been saved (in which
        case this boolean is ``False``).

        Parameters
        ----------
        is_sim_conf_dirty : bool
            ``True`` only if a simulation configuration is currently open *and*
            this configuration is **dirty** (i.e., has unsaved changes).
        '''

        # Set the modification state of this window to correspond to the
        # modification state of this simulation configuration, an operation
        # that has platform-specific effects usually including appending an
        # asterisk to the current window title.
        self.setWindowModified(is_sim_conf_dirty)

    # ..................{ GETTERS                           }..................
    @type_check
    def get_widget(self, widget_name: str) -> QWidget:
        '''
        Widget with the passed name directly owned by this main window if this
        widget exists *or* raise an exception otherwise.

        The Qt (Creator|Designer)-managed ``.ui`` file underlying this window
        declares most of this application's widgets as public instance
        variables of this window, whose variable names are these widget's
        Qt-specific object names. This function provides dynamica access to
        these windows in a safe manner raising human-readable exceptions.

        Caveats
        ----------
        This function is principally intended for use cases in which this
        widget is known only at runtime. Widgets known at development time may
        instead be statically retrieved as public instance variables of this
        window.

        Parameters
        ----------
        widget_name : str
            Name of the widget to be retrieved.

        Returns
        ----------
        QWidget
            This widget.

        Raises
        ----------
        BetseePySideWindowException
            If this window directly owns no such widget.
        '''

        # Widget with this name if any *OR* "None" otherwise.
        widget = self.get_widget_or_none(widget_name)

        # If no such widget exists, raise an exception.
        if widget is None:
            raise BetseePySideWindowException(
                'Qt (Creator|Designer)-managed widget "{}" not found.'.format(
                    widget_name))

        # Return this widget.
        return widget


    @type_check
    def get_widget_or_none(self, widget_name: str) -> QWidgetOrNoneTypes:
        '''
        Widget with the passed name directly owned by this main window if this
        widget exists *or* ``None`` otherwise.

        Parameters
        ----------
        widget_name : str
            Name of the widget to be retrieved.

        Returns
        ----------
        QWidgetOrNoneTypes
            This widget if any *or* ``None`` otherwise.

        See Also
        ----------
        :func:`get_widget`
            Further details.
        '''

        # All that glitters is not gold.
        return getattr(self, widget_name, None)

    # ..................{ RESIZERS                          }..................
    def resize_full(self) -> None:
        '''
        Enable full-screen mode for this main window.

        Specifically, this method:

        * Hides all standard window decorations, including the window frame and
          title bar.
        * Resize this window to consume all available screen space.
        '''

        self.setWindowState(self.windowState() | Qt.WindowFullScreen)


    def resize_max(self) -> None:
        '''
        Resize this window to consume all available screen space.

        Note that:

        * This does *not* constitute "full screen" mode, which is typically
          desirable only for console-oriented applications (e.g., OpenGL).
        * This is *not* supported by Qt Creator and hence must be performed
          with Python logic here.
        * The more convenient :meth:`showMaximized` method is intentionally
          *not* called here, as doing so would implicitly render the entire GUI
          (which could currently be under construction) visible.
        '''

        self.setWindowState(self.windowState() | Qt.WindowMaximized)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all editable simulation configuration widget subclasses
instantiated in pages of the top-level stack.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, Slot
from betse.lib.yaml.abc.yamlabc import YamlABCOrNoneTypes
from betse.lib.yaml.yamlalias import YamlAliasABC
from betse.util.io.log import logs
from betse.util.type.descriptor.datadescs import DataDescriptorBound
from betse.util.type.types import type_check, ClassOrNoneTypes
from betsee.guiexception import BetseePySideWidgetException
from betsee.util.widget.abc.guiwdgabc import QBetseeEditWidgetMixin

# ....................{ MIXINS                             }....................
# To avoid metaclass conflicts with the "QWidget" base class inherited by all
# widgets also inheriting this base class, this base class *CANNOT* be
# associated with another metaclass (e.g., "abc.ABCMeta").
class QBetseeSimConfEditWidgetMixin(QBetseeEditWidgetMixin):
    '''
    Abstract base class of all **editable simulation configuration widget**
    (i.e., widget interactively editing simulation configuration values stored
    in external YAML files) subclasses.

    Design
    ----------
    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be subclassed *first* rather than *last* in subclasses.

    Attributes
    ----------
    _sim_conf : QBetseeSimConf
        High-level state of the currently open simulation configuration, which
        depends on the state of this low-level simulation configuration widget.
    _sim_conf_alias : DataDescriptorBound
        high-level object wrapping the low-level data descriptor of the
        :class:`betse.science.parameters.Parameters` class, itself wrapping the
        lower-level simulation configuration option edited by this widget.
    _sim_conf_alias_type : ClassOrNoneTypes
        Class or tuple of classes that the value to which
        :attr:`_sim_conf_alias` evaluates is required to be an instance of if
        any *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._sim_conf = None
        self._sim_conf_alias = None
        self._sim_conf_alias_type = None


    @type_check
    def init(
        self,

        # Mandatory parameters.
        #
        # To avoid circularity from the "QBetseeSimConf" class importing the
        # "QBetseeMainWindowConfig" class importing the "betsee_ui" submodule
        # importing instances of this class, this type is validated dynamically.
        sim_conf: 'betsee.gui.simconf.guisimconf.QBetseeSimConf',
        sim_conf_alias: YamlAliasABC,

        # Optional parameters.
        sim_conf_alias_parent: YamlABCOrNoneTypes = None,
    ) -> None:
        '''
        Finalize the initialization of this widget.

        Parameters
        ----------
        sim_conf : QBetseeSimConf
            High-level state of the currently open simulation configuration.
        sim_conf_alias : YamlAliasABC
            Low-level data descriptor bound to the simulation configuration
            option edited by this widget, typically a
            :class:`betse.science.params.Parameters`-specific class variable
            assigned the return value of the
            :func:`betse.science.config.confabc.yaml_alias` function.
        sim_conf_alias_parent : YamlABCOrNoneTypes
            YAML-backed simulation subconfiguration whose class declares the
            passed data descriptor. Defaults to ``None``, in which case this
            parameter defaults to ``sim_conf.p`` (i.e., the top-level
            YAML-backed simulation configuration).

        See Also
        ----------
        :meth:`QBetseeWidgetMixinSimConf.init`
            Further details.
        '''

        # Initialize our superclass.
        super().init()

        # Log this initialization *AFTER* storing this name.
        logs.log_debug(
            'Initializing editable widget "%s"...', self.obj_name)

        # If no parent simulation subconfiguration was passed, default to the
        # the top-level simulation configuration.
        if sim_conf_alias_parent is None:
            sim_conf_alias_parent = sim_conf.p

        # Set the undo stack to which this widget pushes undo commands *BEFORE*
        # connecting signals to slots pushing such commands.
        self._set_undo_stack(sim_conf.undo_stack)

        # Classify all passed parameters. Since the main window rather than this
        # state object owns this widget, retaining a reference to this state
        # object introduces no circularity and hence is safe. (That's nice.)
        self._sim_conf = sim_conf

        # Wrap the passed low-level data descriptor with a high-level wrapper
        # bound to the "Parameters" instance associated with this GUI.
        self._sim_conf_alias = DataDescriptorBound(
            obj=sim_conf_alias_parent, data_desc=sim_conf_alias)

        # Type(s) required by this data descriptor if any or "None" otherwise.
        self._sim_conf_alias_type = sim_conf_alias.expr_alias_cls

        # If these type(s) are unsupported by this subclass, raise an exception.
        self._die_if_sim_conf_alias_type_invalid()

        # Populate this widget when opening a simulation configuration.
        self._sim_conf.set_filename_signal.connect(self._set_filename)

    # ..................{ SUBCLASS ~ optional                }..................
    # Subclasses may optionally reimplement the following methods.

    @property
    def _sim_conf_alias_type_strict(self) -> ClassOrNoneTypes:
        '''
        Superclass of all low-level values defined by the simulation
        configuration associated with this widget if this widget is strictly
        constrained to values of only a single type *or* ``None`` otherwise
        (i.e., if this widget loosely displays values satisfying any one of
        several different types).

        By default, this method returns ``None``.

        Caveats
        ----------
        This low-level type is simulation configuration-specific and hence
        distinct from the high-level type(s) of all :mod:`PySide2`-specific
        scalar values displayed by this widget. See the
        :meth:`QBetseeSimConfEditScalarWidgetMixin.widget_value` property for
        discussion of the distinction between these two related types.
        '''

        return None


    def _die_if_sim_conf_alias_type_invalid(self) -> None:
        '''
        Raise an exception if the types of low-level values exposed by the
        simulation configuration alias associated with this widget are
        unsupported by this widget's subclass.
        '''

        # If this widget is *NOT* constrained to values of only a single type,
        # silently noop. In this case, the subclass is required to manually
        # validate these values.
        if self._sim_conf_alias_type_strict is None:
            return
        # Else, this widget is constrained to values of only a single type.

        # True only if this widget is incompatible with this alias.
        is_sim_conf_alias_type_invalid = None

        # This widget is incompatible with this alias if the latter accepts
        # either:
        #
        # * Multiple types excluding this single type.
        # * A single type incompatible with this single type.
        if self._sim_conf_alias_type is tuple:
            is_sim_conf_alias_type_invalid = (
                self._sim_conf_alias_type_strict not in
                self._sim_conf_alias_type)
        else:
            is_sim_conf_alias_type_invalid = not issubclass(
                self._sim_conf_alias_type,
                self._sim_conf_alias_type_strict)

        # If this widget is incompatible with this alias, raise an exception.
        if is_sim_conf_alias_type_invalid:
            raise BetseePySideWidgetException(QCoreApplication.translate(
                'QBetseeSimConfEditWidgetMixin',
                'Widget "{0}" YAML alias type {1!r} != {2!r} '
                '(i.e., expected a type compatible with {2!r} '
                'but received an incompatible type {1!r}).'.format(
                    self.obj_name,
                    self._sim_conf_alias_type,
                    self._sim_conf_alias_type_strict,)))

    # ..................{ PROPERTIES                         }..................
    @property
    def _is_sim_open(self) -> bool:
        '''
        ``True`` only if a simulation configuration file is currently open.
        '''

        return self._sim_conf is not None and self._sim_conf.is_open

    # ..................{ SLOTS                              }..................
    @Slot(str)
    def _set_filename(self, filename: str) -> None:
        '''
        Slot signalled on both the opening of a new simulation configuration
        and closing of an open simulation configuration.

        Design
        ----------
        Subclasses are recommended to override this method by (in order):

        #. Calling this superclass implementation.
        #. If this filename is non-empty, populating this widget's contents with
           the current value of the simulation configuration alias associated
           with this widget.

        Parameters
        ----------
        filename : str
            Absolute path of the currently open YAML-formatted simulation
            configuration file if any *or* the empty string otherwise (i.e., if
            no such file is open).
        '''

        # Undo commands are safely pushable from this widget *ONLY* if a
        # simulation configuration is currently open.
        self.is_undo_cmd_pushable = bool(filename)

    # ..................{ ENABLERS                           }..................
    def _update_sim_conf_dirty(self) -> None:
        '''
        Update the dirty state for the current simulation configuration if this
        widget has been initialized with such a configuration *or* noop
        otherwise.

        This method is intended to be called by subclass slots on completion of
        user edits to the contents of this widget. In response, this method
        notifies all connected slots that this simulation configuration has
        received new unsaved changes.
        '''

        # If this widget has yet to be initialized, silently noop.
        if self._sim_conf is None:
            return

        # "True" only if the undo stack is in the dirty state.
        is_dirty = self._is_undo_stack_dirty()

        # Log this initialization *AFTER* storing this name.
        logs.log_debug(
            'Setting simulation configuration dirty bit from '
            'editable widget "%s" to "%r"...', self.obj_name, is_dirty)

        # Update the dirty state for this simulation configuration.
        self._sim_conf.set_dirty_signal.emit(is_dirty)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **tabbed simulation results** (i.e., partitioning of the
simulation results into multiple pages, each displaying and controlling all
settings associated with a single result of the current simulation) facilities.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import QCoreApplication, QObject, Signal, Slot
from PySide2.QtWidgets import QMainWindow, QTabWidget
# from betse.util.io.log import logs
from betse.util.type.types import type_check  #, StrOrNoneTypes
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ CLASSES                           }....................
class QBetseeSimmerTabWidget(QBetseeObjectMixin, QTabWidget):
    '''
    :mod:`PySide2`-based tab widget containing multiple tabs, each displaying
    and controlling all settings associated with a single simulation result
    (e.g., pickled file, plot, animation) of the current simulation created by
    a single CLI-oriented simulation subcommand (e.g., ``betse plot init``).

    Attributes (Public)
    ----------
    simmer : QBetseeSimmer
        **Simulator** (i.e., :mod:`PySide2`-based object both displaying *and*
        controlling the execution of simulation-specific subcommands).

    Attributes (Private: Non-widgets)
    ----------

    Attributes (Private: Widgets)
    ----------
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulator.
        '''

        # Avoid circular import dependencies.
        from betsee.gui.simtab.run.guisimrun import QBetseeSimmer

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Simulator, displaying and controlling simulation-specific subcommands.
        self.simmer = QBetseeSimmer()

        # Nullify all remaining instance variables for safety.
        # self.simmer = None


    # To avoid circular import dependencies, this parameter is validated to be
    # an instance of the "QMainWindow" superclass rather than the expected
    # "QMainWindow" subclass of the "betsee.gui.window.guimainwindow"
    # submodule. Why? Because the latter imports the cached "betsee_ui.py"
    # module which imports the current submodule. Since this application only
    # contains one main window, this current validation suffices.
    @type_check
    def init(self, main_window: QMainWindow) -> None:
        '''
        Finalize this widget's initialization, owned by the passed main window
        widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulation subcommander.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().init()

        # Initialize all widgets concerning simulation subcommand state the
        # *BEFORE* connecting all relevant signals and slots typically
        # expecting these widgets to be initialized.
        self._init_widgets(main_window)
        self._init_connections(main_window)


    @type_check
    def _init_widgets(self, main_window: QMainWindow) -> None:
        '''
        Create all widgets owned directly by this object *and* initialize all
        other widgets (not necessarily owned by this object) whose internal
        state pertains to the high-level state of simulation subcommands.

        Parameters
        ----------
        main_window : QMainWindow
            Initialized parent :class:`QMainWindow` widget.
        '''

        # Classify all instance variables of this main window subsequently
        # required by this object.
        # self._action_make_sim     = main_window.action_make_sim

        # Initialize the simulator.
        self.simmer.init(main_window=main_window)


    @type_check
    def _init_connections(self, main_window: QMainWindow) -> None:
        '''
        Connect all relevant signals and slots of *all* widgets (including the
        main window, top-level widgets of that window, and leaf widgets
        distributed throughout this application) whose internal state pertains
        to the high-level state of simulation subcommands.

        Parameters
        ----------
        main_window : QMainWindow
            Initialized parent :class:`QMainWindow` widget.
        '''

        # Connect each such action to this object's corresponding slot.
        # self._action_make_sim.triggered.connect(self._make_sim)

        # Connect this object's signals to all corresponding slots.
        # self.set_filename_signal.connect(self.set_filename)

        # Set the state of all widgets dependent upon this simulation
        # subcommand state *AFTER* connecting all relavant signals and slots.
        # Initially, no simulation subcommands have yet to be queued or run.
        #
        # Note that, as this slot only accepts strings, the empty string rather
        # than "None" is intentionally passed for safety.
        # self.set_filename_signal.emit('')

        pass

    # ..................{ FINALIZERS                        }..................
    # Although the implementation of this method is currently trivial, this
    # method will be generalized (at some later date) to transparently halt
    # work being asynchronously performed in additional tabs (e.g., animation
    # encoding export) and thus should be preserved as is for now.
    def halt_work(self) -> None:
        '''
        Schedule all currently running simulation work if any for immediate
        and thus possibly non-graceful termination *or* silently reduce to a
        noop otherwise (i.e., if no simulation work is currently running).

        Caveats
        ----------
        This method may induce data loss or corruption in simulation output.
        In theory, this should only occur in edge cases in which the current
        simulator worker fails to gracefully stop within a sensible window of
        time. In practice, this implies that this method should *only* be
        called when otherwise unavoidable (e.g., at application shutdown).

        See Also
        ----------
        :meth:`QBetseeSimmer.halt_work`
            Further details.
        '''

        # Halt the simulator if currently running.
        self.simmer.halt_work()

    # ..................{ PROPERTIES ~ bool                 }..................
    # @property
    # def is_open(self) -> bool:
    #     '''
    #     ``True`` only if a simulation configuration file is currently open.
    #     '''
    #
    #     return self.p.is_loaded

    # ..................{ PROPERTIES ~ str                  }..................
    # @property
    # def dirname(self) -> StrOrNoneTypes:
    #     '''
    #     Absolute path of the directory containing the currently open
    #     simulation configuration file if any *or* ``None`` otherwise.
    #     '''
    #
    #     return self.p.conf_dirname

    # ..................{ EXCEPTIONS                        }..................
    # def die_unless_open(self) -> bool:
    #     '''
    #     Raise an exception unless a simulation configuration file is currently
    #     open.
    #     '''
    #
    #     if not self.is_open:
    #         raise BetseeSimConfException(
    #             'No simulation configuration currently open.')

    # ..................{ SIGNALS                           }..................
    # set_filename_signal = Signal(str)
    # '''
    # Signal passed either the absolute path of the currently open YAML-formatted
    # simulation configuration file if any *or* the empty string otherwise.
    #
    # This signal is typically emitted on the user:
    #
    # * Opening a new simulation configuration.
    # * Closing a currently open simulation configuration.
    # '''

    # ..................{ SLOTS ~ state                     }..................
    # @Slot(str)
    # def set_filename(self, filename: str) -> None:
    #     '''
    #     Slot signalled on both the opening of a new simulation configuration
    #     and closing of an open simulation configuration.
    #
    #     Parameters
    #     ----------
    #     filename : StrOrNoneTypes
    #         Absolute path of the currently open YAML-formatted simulation
    #         configuration file if any *or* the empty string otherwise (i.e., if
    #         no such file is open).
    #     '''
    #
    #     # Notify all interested slots that no unsaved changes remain, regardless
    #     # of whether a simulation configuration has just been opened or closed.
    #     self.set_dirty_signal.emit(False)

    # ..................{ SLOTS ~ action                    }..................
    # @Slot()
    # def _open_sim(self) -> None:
    #     '''
    #     Slot invoked on the user requesting that the currently open simulation
    #     configuration if any be closed and an existing external simulation
    #     configuration be opened.
    #     '''
    #
    #     # Absolute path of an existing YAML-formatted simulation configuration
    #     # file selected by the user.
    #     conf_filename = self._show_dialog_sim_conf_open()
    #
    #     # If the user canceled this dialog, silently noop.
    #     if conf_filename is None:
    #         return
    #     # Else, the user did *NOT* cancel this dialog.
    #
    #     # Close the currently open simulation configuration if any.
    #     self._close_sim()
    #
    #     # Deserialize this low-level file into a high-level configuration.
    #     self.load(conf_filename)
    #
    #     # Update the status bar *AFTER* successfully completing this action.
    #     self._status_bar.showMessage(QCoreApplication.translate(
    #         'QBetseeSimConf', 'Simulation opened.'))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Core :mod:`PySide2`-specific **type** (i.e., class) functionality.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import (
    QAbstractEventDispatcher,
    QThread,
    QThreadPool,
    Slot,
)
from PySide2.QtWidgets import (
    QLabel,
    QProgressBar,
    QWidget,
)
from betse.util.type.types import NoneType
from betsee.util.widget.stock.guiprogressbar import QBetseeProgressBar

# ....................{ TYPES                             }....................
QWidgetType = QWidget
'''
Root :mod:`PySide2`-based widget type.

This type is synonymous with the stock :class:`QWidget` class and is provided
merely as a caller convenience.
'''

# ....................{ TUPLES ~ none                     }....................
# Tuples of types containing at least the type of the singleton "None" object.

QAbstractEventDispatcherOrNoneTypes = (QAbstractEventDispatcher, NoneType)
'''
Tuple of both the :mod:`PySide2`-based event dispatcher type *and* the type of
the singleton ``None`` object.
'''


SlotOrNoneTypes = (Slot, NoneType)
'''
Tuple of both the :mod:`PySide2`-based slot type *and* the type of the
singleton ``None`` object.
'''

# ....................{ TUPLES ~ none : thread            }....................
QThreadOrNoneTypes = (QThread, NoneType)
'''
Tuple of both the :mod:`PySide2`-based thread type *and* the type of the
singleton ``None`` object.
'''


QThreadPoolOrNoneTypes = (QThreadPool, NoneType)
'''
Tuple of both the :mod:`PySide2`-based thread pool type *and* the type of the
singleton ``None`` object.
'''

# ....................{ TUPLES ~ none : widget            }....................
QLabelOrNoneTypes = (QLabel, NoneType)
'''
Tuple of both the :mod:`Label` widget type *and* the type of the singleton
``None`` object.
'''


QProgressBarOrNoneTypes = (QProgressBar, NoneType)
'''
Tuple of both the :mod:`QProgressBar` widget type *and* the type of the
singleton ``None`` object.
'''


QWidgetOrNoneTypes = (QWidgetType, NoneType)
'''
Tuple of both the root :mod:`PySide2`-based widget type *and* the type of the
singleton ``None`` object.
'''

# ....................{ TUPLES ~ none : widget : betsee   }....................
QBetseeProgressBarOrNoneTypes = (QBetseeProgressBar, NoneType)
'''
Tuple of both the :mod:`QBetseeProgressBar` widget type *and* the type of the
singleton ``None`` object.
'''



#!/usr/bin/env python3
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Root-level classes defining this application's graphical user interface (GUI).
'''

#FIXME: Shutdown gracefully. Currently, shutting the application down tends to
#yield non-deterministic segfaults. Note that properly implementing a graceful
#shutdown procedure is surprisingly non-trivial. As a useful starting point,
#consider the shutdown() and _shutdown() methods of the following submodule:
#    https://github.com/qutebrowser/qutebrowser/blob/master/qutebrowser/app.py
#
#In particular, note the need to:
#
#* Uninstall all event filters and message handlers (i.e., reset these
#  callbacks to their defaults).
#* Use one-shot timers to increase the likelihood of a graceful shutdown.

#FIXME: Dynamically validate the importability of *ALL* PySide2 C extensions
#required by this application. Currently, we do so with simplistic exception
#handling. However, upstream recently noted that:
#
#    - added a way to dynamically detect the available Qt modules in Pyside
#
#What, exactly, is this "way"? If undocumented, let's file an issue requesting
#clarification. First grep the issue tracker for "detect module".

#FIXME: To simplify future localization efforts, all human-readable strings to
#be displayed should be filtered through the Qt tr() or translate() methods.
#Note that the external "pyside2-lupdate" command will probably need to be
#called to convert raw translation files into importable Python modules. For
#further details, see the following PySide-specific articles:
#
#* "Internationalisation of PyQt5 Applications." (Superb article.
#  Unsurprisingly, this is the canonical resource for PySide translations, which
#  conforms to the exact same API and hence suffers similar issues.)
#  http://pyqt.sourceforge.net/Docs/PyQt5/i18n.html
#* "Internationalizing PySide programs." (Crude ELI5 on PySide translations.
#  Insufficient in many respects, but a tolerable introduction to the subject.)
#  https://wiki.qt.io/PySide_Internationalization
#
#Generic Qt articles pertaining to this topic additionally include:
#
#* "Writing Source Code for Translation." (Superb "Qt Assistant" article.)
#* "Qt Linguist Manual: Translators." (Useful "Qt Linguist" usage instructions.)
#  https://doc.qt.io/qt-5/linguist-translators.html
#FIXME: Sadly, the behaviour of the "pyside2-lupdate" command is fairly awkward.
#Unlike the remainder of the PySide2 ecosystem, this command requires a ".pro"
#file whose syntax appears to superficially resemble that of traditional *nix
#makefiles. In turn, this file requires the absolute or relative path to the
#Python... *WAIT.* No, all is well. The ".pro" support is largely vestigial and
#only required for projects *NOT* already providing one or more ".ui" files.
#Fortunately, "pyside2-lupdate" supports both. *WAIT.* O.K.; so, we'll have
#translatable strings residing in:
#
#* Our ".ui" files. These files are guaranteed to be convertible into ".ts"
#  files which may then be translated via the "Qt Linguist" GUI, like so:
#
#    pyside2-lupdate \
#        ~/py/betsee/data/ui/betsee.ui -ts \
#        ~/py/betsee/data/ts/en_US.ts \
#        ~/py/betsee/data/ts/ru_RU.ts   # ...and so on for all needed languages.
#
#* All ".py" files under the ~/py/betsee/betsee/gui/widget/ subdirectory that
#  may contain dynamic calls to the tr() and translate() methods. In theory,
#  these files should also be convertible into ".ts" files which may then be
#  translated via the "Qt Linguist" GUI by recursively finding all such files in
#  pure-Python and then passing the absolute paths of these files in the same
#  manner:
#
#    pyside2-lupdate \
#        ~/py/betsee/betsee/gui/widget/guimainwindow.py \
#        ~/py/betsee/betsee/gui/widget/guisimconftree.py \
#        # ...and so on, recursively.
#        -ts \
#        ~/py/betsee/data/ts/en_US.ts \
#        ~/py/betsee/data/ts/ru_RU.ts   # ...and so on for all needed languages.
#
#  If passing ".py" files directly to "pyside2-lupdate" fails, we'll then need
#  to dynamically generate (and probably cache) a ".pro" file listing the paths
#  of these files, which may then be passed to "pyside2-lupdate" directly. See
#  "pyside2-lupdate --help" for a terse usage synopsis.
#
#Note that ".ts" files are *NOT* usable as is by Qt and hence by PySide2. Such
#files are only intended to be consumed and modified by the "Qt Linguist" GUI.
#Qt and hence PySide2 requires ".ts" files be converted to corresponding ".qm"
#(i.e., [Q]t [M]essage) files by either:
#
#* Manually selecting "File" -> "Release" in the "Qt Linguist" GUI.
#* Programmatically running the "lrelease" command.
#
#Naturally, we overwhelmingly prefer the latter approach -- or we would, at
#least, if the "lrelease" command were distributed as a mandatory core Qt
#component. It's not, which admittedly makes sense. While most Linux
#distributions do provide "lrelease" packages, there's little sense in adding
#yet another runtime dependency to this application... Hmm. Or maybe there is?
#
#O.K., I fully admit: I strongly prefer programmatically running the "lrelease"
#command at runtime and caching the resulting ".qm" files into
#"~/.betse/betsee/qm/" (e.g., "~/.betse/betsee/qm/en_US.qm"). Since translations
#are technically optional, we should simply make "lrelease" an optional
#dependency. Specifically:
#
#* If the "lrelease" command is pathable, the ".qm" target file specific to the
#  current system local should be automatically generated from the corresponding
#  ".ts" source file and cached into the "~/.betse/betsee/qm/" subdirectory.
#* Else, translations are ignored and a non-fatal warning should be issued.
#
#Assuming ".qm" files to be generated and cached in this fashion, we should then
#be able to leverage application-wide translations by adding the following
#initialization functionality to our existing "betsee.util.psdapp" submodule:
#
#    from PySide2.QtCore import QLocale, QTranslator
#
#    app_translator = QTranslator()
#
#    # Obviously, replace "~/.betse/betsee/qm" with the corresponding metadata.
#    #
#    # Perhaps more importantly, this should be strictly optional. Hence, any
#    # exception raised by this call should be converted into a non-fatal logged
#    # warning or (probably) error message. Does Qt provide a means beforehand
#    # of testing whether this ".qm" file exists? I suppose we can just manually
#    # concatenate this file's absolute path ourselves and test for the
#    # existance of this file. Hey, whatever works -- right?
#    app_translator.load(QLocale.system().name(), "~/.betse/betsee/qm")
#
#    get_app().installTranslator(app_translator)
#
#There exists one last minefield: the QObject.tr() versus
#QCoreApplication::translate() split. Basically:
#
#* The QObject.tr() instance method provides a more convenient interface that
#  (unfortunately) is broken by design.
#* QCoreApplication::translate() class method provides a less convenient
#  interface that actually behaves as intended.
#
#Riverbank themselves strongly recommend that QCoreApplication::translate()
#rather than QObject.tr() *ALWAYS* be called. For example, see
#http://pyqt.sourceforge.net/Docs/PyQt5/i18n.html:
#
#"The PyQt5 behaviour is unsatisfactory and may be changed in the future. It is
# recommended that translate() be used in preference to tr(). This is guaranteed
# to work with current and future versions of PyQt5 and makes it much easier to
# share message files between Python and C++ code."
#
#PySide2 is, of course, no different. To compound matters, the "pyside2-lupdate"
#command appears to have difficulty properly associating QObject.tr() calls with
#their appropriate context. For example, see this StackOverflow response:
#    https://stackoverflow.com/a/41174550/2809027
#
#One final caveat exists: interpolated arguments should be explicitly rather
#than implicitly numbered (e.g., "{0}" rather than "{}"). Why? Because
#non-English languages may interpolate arguments in a different order, in which
#case implicitly numbered {}-style arguments would not be translatable. For a
#real-world example demonstrating this difference between two English and
#Norwegian sentences, see the "Using Numbered Arguments" subsection of the
#"Writing Source Code for Translation" Qt article.
#
#Lastly, note that our application would still benefit from a way of permitting
#the user to explicitly define their interface language of choice. Technically,
#automatic locale detection should suffice for now. The appropriate
#documentation reads:
#
#    "Qt notifies applications if the user interface language changes by sending
#     an event of the type QEvent::LanguageChange. To call the member function
#     retranslateUi() of the user interface object, we reimplement
#     QWidget::changeEvent() in the form class, as follows:
#
#     void CalculatorForm::changeEvent(QEvent *e)
#     {
#         QWidget::changeEvent(e);
#         switch (e->type()) {
#         case QEvent::LanguageChange:
#             ui->retranslateUi(this);
#             break;
#         default:
#             break;
#        }
#     }
#
#tl;dr
#=====
#Always call QCoreApplication::translate() and explicitly number interpolated
#arguments (e.g., "{0}" rather than "{}"). To do so, a minimal-length example:
#
#    import sys
#    from QtCore import QCoreApplication
#
#    class MuhObject(QtCore.QObject):
#        def hello(self):
#            return QCoreApplication.translate(
#                'MuhObject', 'Muh hello world, {0}!'.format(sys.argv[0])
#
#Hence, calling QCoreApplication.translate() requires that the first parameter
#be the explicit name of the desired context -- which, for parity with C++,
#should *ALWAYS* be the name of the custom class performing this call.
#
#Annnnnnd we are done.

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from betse.util.io.log import logs
from betse.util.type.types import type_check, StrOrNoneTypes
from betsee import guimetadata
from betsee.gui import guicache
from betsee.util.app import guiapp, guiappstatus, guiappwindow
from betsee.util.io import guierr

# ....................{ CLASSES                            }....................
class BetseeGUI(object):
    '''
    Graphical user interface (GUI) for this application, doubling as both the
    main window and root Qt widget for this application.

    Attributes
    ----------
    _settings : QBetseeSettings
        :class:`QSettings`-based object exposing all application-wide settings
        via cross-platform, thread- and process-safe slots.
    _signaler : QBetseeSignaler
        :class:`PySide2`-based collection of all application-wide signals.
    _sim_conf_filename : StrOrNoneTypes
        Absolute or relative path of the initial YAML-formatted simulation
        configuration file to be initially opened if any *or* ``None``
        otherwise.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, sim_conf_filename: StrOrNoneTypes) -> None:
        '''
        Initialize this graphical user interface (GUI).

        Parameters
        ----------
        sim_conf_filename : StrOrNoneTypes
            Absolute or relative path of the initial YAML-formatted simulation
            configuration file to be initially opened if any *or* ``None``
            otherwise.
        '''

        # Initialize subclasses performing diamond inheritance if any.
        super().__init__()

        # Classify all passed parameters.
        self._sim_conf_filename = sim_conf_filename

        # Nullify all instance variables for safety.
        self._signaler = None
        self._settings = None

        # Generate all modules required at runtime by this GUI.
        guicache.cache_py_files()

    # ..................{ RUNNERS                            }..................
    def run(self) -> int:
        '''
        Run this GUI's main event loop and display this GUI.

        Returns
        ----------
        int
            Exit status of this event loop as an unsigned byte.
        '''

        # Install a global exception hook overriding PySide2's default insane
        # exception handling behaviour with sane exception handling *BEFORE*
        # triggering any application-specific slots possibly raising exceptions.
        # Since the _make_main_window() method transitively does so, this hook
        # should be installed as the first call of this method.
        guierr.install_exception_hook()

        # Create but do *NOT* display this GUI's main window.
        self._make_main_window()

        # Run this GUI's main event loop and display this GUI.
        return self._show_main_window()


    def _make_main_window(self) -> None:
        '''
        Create but do *not* display this GUI's main window.
        '''

        # Since the main window Qt widget class for this application class
        # subclasses the custom user interface (UI) base class defined by a
        # module generated at runtime above, this importation is deferred until
        # *AFTER* this module is guaranteed to be importable.
        from betsee.gui.guisettings import QBetseeSettings
        from betsee.gui.guisignal import QBetseeSignaler
        from betsee.gui.window.guimainwindow import QBetseeMainWindow

        # Log this initialization.
        logs.log_info('Initiating PySide2 UI...')

        # Application-wide signaler.
        self._signaler = QBetseeSignaler()

        # Main window widget for this GUI, which requires this signaler.
        #
        # For safety, this window is scoped to an instance rather than global
        # variable, ensuring this window is destroyed *BEFORE* the root Qt
        # application widget containing this window.
        #
        # To permit external callers throughout the codebase to access this
        # variable, this variable is exposed as a public attribute of the
        # singleton application widget rather than this less accessible object.
        # See the "guiapp" submodule for further details.
        main_window = QBetseeMainWindow(
            signaler=self._signaler,
            sim_conf_filename=self._sim_conf_filename,
        )
        # logs.log_debug('Main window parent: %r', main_window.parentWidget())

        # Publicize this main window for use elsewhere in the codebase.
        guiappwindow.set_main_window(main_window)

        # Display a sensible startup message in this window's status bar *AFTER*
        # publicizing this window, as the former requires the latter.
        guiappstatus.show_status(QCoreApplication.translate(
            'BetseeGUI', 'Welcome to {}'.format(guimetadata.NAME)))

        # Application-wide settings slotter, which requires this window.
        self._settings = QBetseeSettings(main_window)

        # Restore previously stored application-wide settings *AFTER*
        # initializing but *BEFORE* displaying this window.
        self._settings.restore_settings()

        # Finalize this window *AFTER* restoring settings, which modifies
        # critical window properties.
        #
        # Note that this call does *NOT* actually display this window. The main
        # event loop required to do so has yet to be run by the GUI_APP.exec_()
        # call. In this case, this method merely toggles this window's internal
        # state in preparation for that call. (Why, Qt. Why.)
        main_window.show()
        # main_window.showMaximized()


    def _show_main_window(self) -> int:
        '''
        Display this GUI's previously created main window.

        Specifically, this method:

        * Run this GUI's event loop, thus displaying this window.
        * Propagates the resulting exit status to the caller.
        '''

        # Log this display.
        logs.log_info('Displaying PySide2 UI...')

        # Application singleton, guaranteed to have been instantiated.
        gui_app = guiapp.get_app()

        # Display this GUI.
        return gui_app.exec_()
